from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
import uuid


from ...gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DScatterPlotItem, GL3DSurfacePlotItem, GL3DLinePlotDataItem
from ...gl_3D_graphing.graphics_items.GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from ...gl_3D_graphing.graphics_items.GL3DPlotDataItemMixin import BaseGL3DPlotDataItemMixin
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
import traceback
from abc import abstractmethod


class ABCPyVolPlotItemMixin(BaseGL3DPlotDataItemMixin):
    def __init__(self, px_type=None, *args, **kwargs):
        self.px_type = px_type
        super().__init__(*args, **kwargs)
         
    @abstractmethod
    def getValues(self):...
    
    @abstractmethod
    def id(self):...
    
    
class PyVolGL3DSurfacePlotItem(ABCPyVolPlotItemMixin, GL3DSurfacePlotItem):
    def __init__(self, colormap=None, color=None, *args, **kwargs):
        kwargs["shader"] = "shaded"
        self.colormap=colormap
        self.color=color
        self.init=True
        kwargs["color"]=color
        kwargs["colors"]=colormap.map(kwargs["z"]) if not colormap is None else None
        self.valid_values=False
        self.item_type=type
        self._internal_id=uuid.uuid4()
        self.color=kwargs["color"]
        self.original_setData: Callable=None
        self.dataset = None
        
        if isinstance(kwargs["color"], str):
            qcolor =  QtGui.QColor(kwargs["color"])
            normalised_rgba = (qcolor.redF(), qcolor.greenF(), qcolor.blueF(), qcolor.alphaF())
            kwargs["color"] = normalised_rgba
            
        super().__init__(*args, **kwargs)
        self.setGLOptions('opaque')
            
    def setData(self, *args, **kwargs):
        if len(args) >= 4 or "colors" in kwargs:
            return super().setData(*args, **kwargs)
        else:
            if len(args) >= 3:
                z_values = args[2]             
            else:
                z_values = kwargs.get("z")       
            if z_values is not None:
                kwargs["colors"] = self.colormap.map(z_values)
            super().setData(*args, **kwargs)     
            self.valid_values=~np.isnan(self._z).all()
        self.sigPlotChanged.emit(self)

    def id(self):
        return self._internal_id

    def getValues(self):
        return self._x, self._y, self._z
    
    
class PyVolGL3DScatterPlotItem(ABCPyVolPlotItemMixin, GL3DScatterPlotItem):
    def __init__(self, *args, **kwargs):
        kwargs["size"] = 5
        self.item_type = None
        self._internal_id=uuid.uuid4()
        self.color=kwargs["color"]
        self.original_setData: Callable=None
        self.dataset = None
        self.valid_values=False
        
        if isinstance(kwargs["color"], str):
            qcolor =  QtGui.QColor(kwargs["color"])
            normalised_rgba = (qcolor.redF(), qcolor.greenF(), qcolor.blueF(), qcolor.alphaF())
            kwargs["color"] = normalised_rgba

        super().__init__(*args, **kwargs)
        
    def setData(self, *args, **kwargs):
        super().setData(*args, **kwargs)
        self.valid_values=True
        self.sigPlotChanged.emit(self)

    def id(self):
        return self._internal_id
    
    def getValues(self):
        return self.pos[:,0], self.pos[:,1], self.pos[:,2]

class PyVolLineItem(ABCPyVolPlotItemMixin, GL3DLinePlotDataItem):
    def __init__(self, px_type=None, *args, **kwargs):
        self.px_type=px_type
        self.item_type = None
        self._internal_id=uuid.uuid4()
        self.dataset = None
        super().__init__(*args, **kwargs)

    def id(self):
        return self._internal_id
    
    def getValues(self):
        return self.pos[:,0], self.pos[:,1], self.pos[:,2]



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option, ABCInstrument, Derivative
    from instruments.utils import InstrumentManager
    from ..data_classes.classes import Slice
    
import pyqtgraph as pg
import numpy as np
from PySide6 import QtCore, QtWidgets, QtGui
from ..metric_attributes import get_metric_category
from .option_monitor import extra_widgets, stylesheets as utils_omon
from pprint import pprint


class CustomTextItem(pg.TextItem):
    def __init__(self, text=None, view_box_anchor=None, **kwargs):
        self.view_box_anchor=view_box_anchor
        if self.view_box_anchor[0] == 0:
            self.use_vb_xmin=True
        else:
            self.use_vb_xmin=False
        if self.view_box_anchor[1] == 0:
            self.use_vb_ymin=False
        else:
            self.use_vb_ymin=True

        super().__init__(text=text, **kwargs)
        self.p=[self.pos()[0], self.pos()[1]]
    
    def attach_viewbox(self, vb: pg.ViewBox):
        self.vb=vb 
        self.vb.sigResized.connect(self.view_box_resize)
        self.vb.sigRangeChanged.connect(self.view_box_resize)
        self.vb.sigRangeChangedManually.connect(self.view_box_resize)


    def right_click(self, *args):
        if not self.isVisible():
            super().show()
    
    def view_box_resize(self, args):
        if not self.view_box_anchor is None:
            vb_range = self.vb.viewRange()
            if self.use_vb_xmin:
                self.p[0] = vb_range[0][0]
            else:
                self.p[0] = vb_range[0][1]
            if self.use_vb_ymin:
                self.p[1] = vb_range[1][0]
            else:
                self.p[1] = vb_range[1][1]
            super().setPos(*self.p)

class CustomPlotDataItem(pg.PlotDataItem):
    def __init__(self, px_type, *args, **kwargs):
        self.px_type=px_type  
        super().__init__(*args, **kwargs)
    
    def update_from_dataclass(self, dclass: Slice):
        self.setData(**dclass.plot_item_kwargs())
    

class OptionInfoTable(QtWidgets.QWidget):
    def __init__(self, name_metric_map, getattr_map, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getattr_map=getattr_map
        self.price_types=name_metric_map["price_types"]    
        self.cellitem_container={}
        self._on_view=True
        self._initLayout()
        self._initTables(name_metric_map, getattr_map)
        
    def _initLayout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    """
    def _initTables(self, name_metric_map, getattr_map):
        self.name_table = NameTable(name_metric_map["object"].ticker,
                                    name_metric_map["object"].underlying_px)
        
        self.metric_table = MetricTable(name_metric_map, getattr_map)
        self.layout().addWidget(self.name_table)
        self.layout().addWidget(self.metric_table)
        
        self.autoFillBackground()
        self.pop_table=False
        self.metric_table.verticalHeader().sectionResized.connect(self.update_vertical_dims)
        self.name_table.horizontalHeader().sectionResized.connect(self.update_horizontal_dims)
        self.showEvent = self.on_show_event
    """
    def update_horizontal_dims(self, *args):
        self.name_table.horizontalHeader().setFixedWidth(self.metric_table.horizontalHeader().width())
        
    def update_vertical_dims(self, *args):
        self.metric_table.verticalHeader().setFixedWidth(self.metric_table.verticalHeader().width())
        

    def on_show_event(self, event):
        self.name_table.verticalHeader().resizeSections(QtWidgets.QHeaderView.ResizeMode.ResizeToContents) 
        self.update_vertical_dims(0, 0, 0) 
        super().showEvent(event)





from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from PySide6.QtCore import Signal 
    from custom_widgets import CustomPlotDataItem
    from ..data_classes.classes import Slice
    
    
import pyqtgraph as pg
import time
from . import custom_widgets
from PySide6 import QtCore 

title_dataclass_type_map = {"Skew" : "skew",
                            "Term Structure" : "term"}




#%%
from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, ClassVar

from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GLViewBox import GLViewBox
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from pyqtgraph.Qt import QT_LIB
import importlib
import pyqtgraph as pg
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DViewBox
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DViewBox, GLSurfacePlotItem, AbstractGLPlotItem, AbstractGLGraphicsItem, GLScatterPlotItem
from pyqtgraph import opengl
import sys
import cProfile
import sys
import pstats
from pyqtgraph import ButtonItem, icons
from dataclasses import dataclass, InitVar, field
from abc import ABC, abstractmethod
from PySide6 import QtGui, QtCore, QtWidgets
from PySide6.QtWidgets import QGraphicsItem
from enum import Enum, EnumType, Flag, IntFlag
from pyqtgraph import Transform3D
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
import warnings
import weakref


@dataclass(slots=True)
class AbstractPlotDataset(ABC):
    initData: InitVar[np.ndarray|Tuple[np.ndarray, ...]] = None
    allFinite: List[Optional[bool]] = field(default_factory=list)
    _dataBounds: List[List[float]] | None = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._initialize_attributes()
        
        
    def _initialize_attributes(self):
        if len(self.allFinite) == 0:
            self.allFinite = [None] * self.num_components
            self._updateDataBounds()
        
    def dataBounds(self) -> QtCore.QRectF | None:
        if self._dataBounds is None: 
            self._updateDataBounds()
        return self._dataBounds
    
    @abstractmethod
    def data(self) -> np.ndarray | Tuple[np.ndarray,...]:
        ...
        
    @abstractmethod
    def dataKwargs(self) -> Dict[str, np.ndarray]:
        ...


        
    @property
    @abstractmethod
    def num_components(self) -> int:
        """Number of data components (e.g., 3 for x/y/z)"""
        
    @abstractmethod
    def getComponents(self) -> List[np.ndarray]:
        """Get list of component arrays (flattened for bounds calculation)"""
    
    def _updateDataBounds(self):
        components = self.getComponents()
        if any([values is None for values in components]):
            return None
         
        dataBounds=[]
        for idx, arr in enumerate(components):
            minRange, maxRange, finiteFlag = self._getArrayBounds(arr, self.allFinite[idx])
            dataBounds.append((minRange, maxRange))
            self.allFinite[idx]=finiteFlag
        self._dataBounds=dataBounds
                
    def _getArrayBounds(self,
                        arr: np.ndarray,
                        all_finite: bool | None
                        ) -> tuple[float, float, bool]:
        # here all_finite could be [None, False, True]
        if not all_finite:  # This may contain NaN or inf values.
            # We are looking for the bounds of the plottable data set. Infinite and Nan
            # are ignored.
            selection = np.isfinite(arr)
            # True if all values are finite, False if there are any non-finites
            all_finite = bool(selection.all())
            if not all_finite:
                arr = arr[selection]
        
        # here all_finite could be [False, True]
        try:
            amin = np.min( arr )  # find minimum of all finite values
            amax = np.max( arr )  # find maximum of all finite values
        except ValueError:  # is raised when there are no finite values
            amin = np.nan
            amax = np.nan
        return amin, amax, all_finite

@dataclass(slots=True)
class PlotDatasetFlatMesh(AbstractPlotDataset):    
    _pos: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self._pos = initData
        super(PlotDatasetFlatMesh, self).__post_init__(initData)
        
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        return [self._pos[:, i] for i in range(self.num_components)]
    
    @property
    def pos(self) -> np.ndarray:
        return self._pos
        
    @pos.setter
    def pos(self, value: np.ndarray):
        self._pos = value

    def data(self) -> np.ndarray:
        return self.pos
    
    def dataKwargs(self):
        return {"pos" : self.pos}


            
@dataclass(slots=True)
class PlotDatasetMesh(AbstractPlotDataset):
    initData: InitVar[Tuple[np.ndarray, ...]] = None
    
    _x: np.ndarray = field(init=False, default=None)
    _y: np.ndarray = field(init=False, default=None)
    _z: np.ndarray = field(init=False, default=None)
    
    def __post_init__(self, initData):
        self.x, self.y, self.z = initData
        super(PlotDatasetMesh, self).__post_init__(initData)
            
    @property
    def num_components(self) -> int:
        return 3
        
    def getComponents(self) -> List[np.ndarray]:
        return [c.ravel() for c in (self.x, self.y, self.z)]
    
    # Property setters with automatic bounds update
    @property
    def x(self) -> np.ndarray:
        return self._x
        
    @x.setter
    def x(self, value: np.ndarray):
        self._x = value
    
    @property
    def y(self) -> np.ndarray:
        return self._y
        
    @y.setter
    def y(self, value: np.ndarray):
        self._y = value
    
    @property
    def z(self) -> np.ndarray:
        return self._z
        
    @z.setter
    def z(self, value: np.ndarray):
        self._z = value

    def data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return (self.x, self.y, self.z)

    def dataKwargs(self):
        return {"x" : self.x, "y" : self.y, "z" : self.z}

GLOptions = {
    'opaque': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: False,
        GL.GL_CULL_FACE: False,
    },
    'translucent': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFuncSeparate': (GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA,
                                GL.GL_ONE, GL.GL_ONE_MINUS_SRC_ALPHA),
    },
    'additive': {
        GL.GL_DEPTH_TEST: False,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFunc': (GL.GL_SRC_ALPHA, GL.GL_ONE),
    },
}    




qt_enums = [QGraphicsItem.GraphicsItemChange.ItemPositionChange,
            QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleChange, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemSceneChange, 
            QGraphicsItem.GraphicsItemChange.ItemTransformChange,
            QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemParentChange,
            QGraphicsItem.GraphicsItemChange.ItemParentHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemChildAddedChange,
            QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
            ]
    

class GLGraphicsEnumMetaclass(EnumType):
    def __new__(metacls, clsname, bases, clsdict):
        base_enums = clsdict.get('_base_enums', [])
        custom_members = clsdict.get('_custom_members', [])

        ignore_list = list(clsdict.get('_ignore_', []))
        ignore_list.extend(['_base_enums', '_custom_members'])
        clsdict['_ignore_'] = ignore_list

        values = [member.value for member in base_enums]
        max_value = max(values) if values else 0

        for member in base_enums:
            clsdict[member.name] = member.value

        shifts = [1, 2]
        for idx, (name, _) in enumerate(custom_members):
            shift = shifts[idx] if idx < len(shifts) else (idx + 1)
            value = max_value << shift
            clsdict[name] = value

        return super().__new__(metacls, clsname, bases, clsdict)
    
    def __call__(cls, value):
        return super().__call__(value)


qt_flags = [QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations,
            QGraphicsItem.GraphicsItemFlag.ItemHasNoContents,
            QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges,
            ]


class MetaFlag(type(IntFlag)):
    def __call__(cls, value=0):
        return super().__call__(value)




class GLGraphicsItemMixin:
    
    _nextId = 0
    
        
    class GLGraphicsItemChange(Enum, metaclass=GLGraphicsEnumMetaclass):
        _ignore_ = ['_base_enums', '_custom_members'] 
        _base_enums = qt_enums
        _custom_members = [('ItemViewChange', None), ('ItemViewHasChanged', None),
                           ('ItemParentWidgetChange', None), ("ItemParentWidgetHasChanged", None),
          #                 ('ItemDataChange', None), ("ItemParentWidgetHasChanged", None),
                           ]

    class GLGraphicsItemFlag(IntFlag, metaclass=MetaFlag):
        NoFlag = 0
        ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
        ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
        ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value


        def __repr__(self):
            if self.value == 0:
                return f"{self.__class__.__name__}({self.value})"
            return super().__repr__()

    
    wrapped_methods = ["setData", "paint"]
    

    def __init__(self, parentItem: 'GLGraphicsItemMixin'=None, **kwargs):
        
        
        self.__parent = None
        self.__view = None
        self.__children = list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__initialized = False
        self.__glOpts = {}
        self.__flags = self.GLGraphicsItemFlag()

        print("\n")
        print("abcGraphicsItem")
        print(f"cls: {self.__class__.__name__}")
        print(f"bases: {self.__class__.__bases__}")
        print(f"isinstance(self, GLGraphicsItem): {isinstance(self, GLGraphicsItem)}")

            
            
        self._connectedView = None
        self._id = GLGraphicsItemMixin._nextId
        GLGraphicsItemMixin._nextId += 1
        self.__blockUpdates: bool=False
        self.__cachedView=None
        self.__parentWidget=None
        self.blockUpdates(False)
        super().__init__(parentItem=parentItem)
        
        print("last")



        self.__parent: GLGraphicsItemMixin | None = None
        self.__view = None
        self.__children: list[GLGraphicsItemMixin] = []#list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__flags = self.GLGraphicsItemFlag()
        
        self.setFlag(self.GLGraphicsItemFlag.ItemSendsGeometryChanges)

        self.__initialized = False
        self.setParentItem(parentItem)
        self.setDepthValue(0)
        self.__glOpts = {}

                
    def blockUpdates(self, flag):
        self.__blockUpdates=flag

    def setFlag(self, flag, enabled=True):
        if isinstance(flag, Flag):
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsChange, enabled)
            
            if enabled:
                new_flags = self.flags() | flag
            else:
                new_flags = self.flags() & ~flag
            
            if self.flags() != new_flags:
                self.__flags = new_flags
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsHaveChanged, enabled)
    
    def flags(self):    
        return self.__flags

    def setParentItem(self, item: 'GLGraphicsItemMixin'):
        changed_enums=[]
        
        if not self.__parent is None:
            if item == self.__parent:
                print(f"self: {self}")
                print(f"item: {item}")
                warnings.warn(f"{item.__class__.__name__} is already a parentItem to {self.__class__.__name__}, ignoring...", UserWarning)
                return
            else:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                self.itemChange(self.GLGraphicsItemChange.ItemChildRemovedChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
                self.__parent.__removeChild(self)
            
        if not item is None:    
            if len(changed_enums) == 0:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
            changed_enums.append((self.GLGraphicsItemChange.ItemChildAddedChange, self))
            item.__addChild(self)
            
        if not self.__view is None:
            self.itemChange(self.GLGraphicsItemChange.ItemViewChange, item)
            changed_enums.append((self.GLGraphicsItemChange.ItemViewHasChanged, item))
            self.__view.removeItem(self)
        self.__parent = item
        self.__view = None
        
        for change, value in changed_enums:
            self.itemChange(change, value)
    
    
    def __addChild(self, childItem: 'GLGraphicsItemMixin'):
        """Do NOT call this method or override it with setParentItem to the childItem. 
        Doing so will cause infinite recursion."""
        self.__children.append(childItem)

        
    def __removeChild(self, childItem: 'GLGraphicsItemMixin'):
        if childItem in self.__children:
            self.__children.remove(childItem)

    def __paintHelper(self, *args, **kwargs):
        if QGraphicsItem.GraphicsItemFlag.ItemHasNoContents & self.flags():
            return 
        
    def visible(self):
        return self.__visible
        
    def itemChange(self, change, value):
        if change in [self.GLGraphicsItemChange.ItemParentHasChanged, self.GLGraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()

    
    
    def setGLOptions(self, opts):

        if isinstance(opts, str):
            opts = GLOptions[opts]
        self.__glOpts = opts.copy()
        self.update()
        
    def updateGLOptions(self, opts):
        """
        Modify the OpenGL state options to use immediately before drawing this item.
        *opts* must be a dictionary as specified by setGLOptions.
        Values may also be None, in which case the key will be ignored.
        """
        self.__glOpts.update(opts)
        
    
    def parentItem(self):
        """Return a this item's parent in the scenegraph hierarchy."""
        return self.__parent
        
        
    def _setView(self, v):
        self.__view = v
        
    def view(self):
        if self.__parent is None:
            # top level object
            return self.__view
        else:
            # recurse
            return self.__parent.view()
        
    def setDepthValue(self, value):

        self.__depthValue = value
        
    def depthValue(self):
        return self.__depthValue
    
    def __setTransformHelper(self, tr):
        self.itemChange(self.GLGraphicsItemChange.ItemTransformChange, tr)
        self.__transform = tr
        self.itemChange(self.GLGraphicsItemChange.ItemTransformHasChanged, tr)
        self.update()

        
    def setTransform(self, tr):
        if not isinstance(tr, Transform3D):
            tr = Transform3D(tr)
        if tr != self.transform():
            self.__setTransformHelper(tr)
        
    def resetTransform(self):
        if self.transform().isIdentity():
            return 
        self.__setTransformHelper(Transform3D())

        
    def applyTransform(self, tr, local):
        
        if local:
            newTransform = self.transform() * tr
        else:
            newTransform = tr * self.transform()
        if newTransform != self.transform():
            self.__setTransformHelper(newTransform)

            
    def transform(self) -> Transform3D:
        """Return this item's transform object."""
        return self.__transform
        
    def viewTransform(self):
        tr = self.__transform
        p = self
        while True:
            p = p.parentItem()
            if p is None:
                break
            tr = p.transform() * tr
        return Transform3D(tr)
        
    def translate(self, dx, dy, dz, local=False):
        tr = Transform3D()
        tr.translate(dx, dy, dz)
        self.applyTransform(tr, local=local)
        
    def rotate(self, angle, x, y, z, local=False):
        """
        Rotate the object around the axis specified by (x,y,z).
        *angle* is in degrees.
        
        """
        tr = Transform3D()
        tr.rotate(angle, x, y, z)
        self.applyTransform(tr, local=local)
    
    def scale(self, x, y, z, local=True):
        """
        Scale the object by (*dx*, *dy*, *dz*) in its local coordinate system.
        If *local* is False, then scale takes place in the parent's coordinates.
        """
        tr = Transform3D()
        tr.scale(x, y, z)
        self.applyTransform(tr, local=local)
    
    
    def hide(self):
        self.setVisible(False, True)
        
    def show(self):
        self.setVisible(True, True)
    
    
    def initialize(self):
        self.initializeGL()
        self.__initialized = True

    def isInitialized(self):
        return self.__initialized
    
    def initializeGL(self):

        pass
    
    def setupGLState(self):
        for k,v in self.__glOpts.items():
            if v is None:
                continue
            if isinstance(k, str):
                func = getattr(GL, k)
                func(*v)
            else:
                if v is True:
                    GL.glEnable(k)
                else:
                    GL.glDisable(k)
    """
    def paint(self):
        self.setupGLState()
        
    def update(self):
        if self.updatesBlocked():
            return

        v = self.view()
        if v is None:
            return
        v.update()
    """
    
    def updatesBlocked(self):
        return self.__blockUpdates

    def setVisible(self, value, explicitly):
        self.__setVisibleHelper(value, True, explicitly)

    def __setVisibleHelper(self, visible, update, explicitly):
        if explicitly:
            self.__explicitlyHidden = visible 
            
        if not self.parentItem() is None and visible and not self.parentItem().visible():
            return 

        if visible == self.visible():
            return 
            
        newVisible = self.itemChange(self.GLGraphicsItemChange.ItemVisibleChange, visible)
        if self.visible() == newVisible:
            return
        
        self.__visible = newVisible
        
        for child in self.childItems():
            if not visible or not child.__explicitlyHidden:
                child.__setVisibleHelper(visible, False, False)
        
        self.update()
        self.itemChange(self.GLGraphicsItemChange.ItemVisibleHasChanged, self.visible())
    
    def childItems(self) -> List['GLGraphicsItemMixin']:
        return list(self.__children)
    
   
    def mapToParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)
        
    def mapToView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)

    def modelViewMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentModelView()

    def projectionMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection()

    def mvpMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection() * view.currentModelView()


    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return
        self.forgetViewBox()
        
        ## check for this item's current viewbox
        view = self.getViewBox()
        
        oldView = None
        if not self._connectedView is None:
            oldView = self._connectedView()
        
        if view is oldView:
            # "  already have view", view
            return

        ## disconnect from previous view
        if not oldView is None:
            for signal, slot in [('sigRangeChanged', self.viewRangeChanged),
                                 ('sigTransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None
            
        if not view is None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        self._replaceView(oldView)        
        self.viewChanged(view, oldView)        

    def setParentWidget(self, parent=None):
        
        if self.parentWidget() != parent:
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetChange, parent) 
            self.__parentWidget = parent
            self.itemChange(self.GLGraphicsItemChange.ItemParentWidgetHasChanged, parent) 
    
    def parentWidget(self) -> GL3DViewWidget:
        return self.__parentWidget

    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a ViewBox)"""
        
    def viewRect(self):
        """Return the visible bounds of this item's ViewBox or GraphicsWidget,
        in the local coordinate system of the item."""
        if self.__cachedView is not None:
            return self.__cachedView

        # Note that in cases of early returns here, the view cache stays empty (None).
        view = self.getViewBox()
        if view is None:
            return None
        bounds = self.mapRectFromView(view.viewRect())
        if bounds is None:
            return None

        bounds = bounds.normalized()
        
        self.__cachedView = bounds
        
        ## nah.
        #for p in self.getBoundingParents():
            #bounds &= self.mapRectFromScene(p.sceneBoundingRect())

        return bounds
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if isinstance(child, '"GLGraphicsItemMixin"'):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        

    def parentChanged(self):
        self.changeParent(self)
        
    def changeParent(self):
        self._updateView()

    
    @QtCore.Slot()
    def viewTransformChanged(self):
        self.__cachedView = None
        
    
        
    def informViewBoundsChanged(self):
        view = self.getViewBox()
        if view is not None and hasattr(view, 'implements') and view.implements('ViewBox'):
            view.itemBoundsChanged(self)
    

    def allChildItems(self, root=None):
        """Return list of the entire item tree descending from this item."""
        if root is None:
            root = self
        tree = []
        for ch in root.childItems():
            tree.append(ch)
            tree.extend(self.allChildItems(ch))
        return tree
    
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):...


    def mapRectFromView(self, obj):
        vt = self.viewTransform()
        if vt is None:
            return None
        vt = fn.invertQTransform(vt)
        return vt.mapRect(obj)

    def getViewBox2(self):
        

        if self._viewBox is None:
            while True:
                parent = self.parentItem()
                if isinstance(parent, GL3DViewBox):
                    self._viewBox = parent
                    break
            return self._viewBox if self._viewBox else None

    def getViewBox(self) -> GL3DViewBox:
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:  ## sometimes happens as items are being removed from a scene and collected.
                    return None
                if p is None:
                    return None
                if hasattr(type(p), 'implements'):
                    if p.implements('ViewBox'):
                        self._viewBox = weakref.ref(p)
                        break
        
        return self._viewBox() if self._viewBox else None
    
    
    def forgetViewBox(self):
        self._viewBox = None
"""
class _Temp2("GLGraphicsItemMixin"):
    pass

class AbstractMeta(meta.QABCMeta):
    def __call__(cls, *args, **kwargs):
        self = cls.__new__(cls, *args, **kwargs)
        
        self._original_class = cls
        self.__class__ = _Temp
        parentItem = kwargs.pop("parentItem", None)
        "GLGraphicsItemMixin".__init__(self, parentItem=parentItem)
        
        self.__class__ = self._original_class
        del self._original_class
        
        if hasattr(self, '__init__') and self.__init__ is not None:
            self.__init__(*args, **kwargs)
        
        
        return self

"""

from dataclasses import dataclass
class MeshPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._x: np.ndarray
        self._y: np.ndarray 
        self._z: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def x(self):
        return self._x
      
    @x.setter
    def _x(self, value):
        self._x = value
        
    @property
    def y(self):
        return self._y
      
    @y.setter
    def _y(self, value):
        self._y = value

    @property
    def z(self):
        return self._z
      
    @z.setter
    def _z(self, value):
        self._z = value

class VectorPlotitemProperties:
    def __init__(self, *args, **kwargs):
        self._pos: np.ndarray
        super().__init__(*args, **kwargs)
    
    @property
    def pos(self):
        
        return self._pos
      
    @pos.setter
    def _pos(self, value):
        if (self.pos != value).any():
            
            self._pos = value
        
@dataclass(slots=True)
class CacheData:
    _null3: ClassVar[List[None]] = [None, None, None]
    _false3: ClassVar[List[None]] = [False, False, False]
    _null32: ClassVar[List[List[None]]] = [[None, None],[None, None], [None, None]]
    
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _bounds: List[List[None]]|List[List[float]] = field(init=False, default_factory=lambda : [row.copy() for row in CacheData._null32])
    _mask: List[None]|List[np.ndarray] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _orthoRange: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _frac: List[None]|List[float] = field(init=False, default_factory=lambda : CacheData._null3.copy())
    _cacheFlag: List[bool] = field(init=False, default_factory=lambda : CacheData._false3.copy())

    def __repr__(self):
        out = {}
        for name in self.__slots__:
            if name.startswith('_'):
                public_name = name[1:]
                method = getattr(self, public_name, None)
                if callable(method):
                    try:
                        out[public_name] = method()
                    except Exception:
                        out[public_name] = "<error>"
        return f"{self.__class__.__name__}({out})"   
     
    def cacheFlag(self, ax=None):
        if ax is None:
            return self._cacheFlag
        else:
            return self._cacheFlag[ax]
        
    
    def setCacheFlag(self, flag, ax=None):
        if ax is None:
            self._cacheFlag=flag
        else:
            self._cacheFlag[flag] = ax

    def setBounds(self, bounds, ax=None):
        if ax is None:
            self._bounds = bounds
            self.setCacheFlag([not None in lim for lim in bounds])
        else:
            self._bounds[ax] = bounds
            self.setCacheFlag(bounds, ax)

    
    def bounds(self, ax=None):
        if ax is None:
            return self._bounds
        else:
            return self._bounds[ax]

    
    def setMask(self, mask, ax=None):
        if ax is None:
            self._mask=mask     
        else:
            self._mask[ax] = mask
    
        
    def mask(self, ax=None):
        if ax is None:
            self._mask
        else:
            return self._mask[ax]


    def setOrthoRange(self, orthoRange, ax=None):
        if ax is None:
            self._orthoRange=orthoRange
        else:
            self._orthoRange[ax] = orthoRange

    def orthoRange(self, ax=None):
        if ax is None:
            return self._orthoRange
        else:
            return self._orthoRange[ax] 


    def setFrac(self, frac, ax=None):
        if ax is None:
            self._frac=frac
        else:
            self._frac[ax] = frac
        
    def frac(self, ax=None):
        if ax is None:
            return self._frac
        else:
            return self._frac[ax] 

    def reset(self):
        self.setBounds([row.copy() for row in self._null32])
        self.setCacheFlag(self._false3.copy())
        self.setMask(self._null3.copy())
        self.setOrthoRange(self._null3.copy())
        self.setFrac(self._null3.copy())




class AbstractGLPlotItem(GLGraphicsItemMixin):
    sigPlotChanged = QtCore.Signal(object) 
    
    def __init__(self, topology_type="", view_box=None, parentItem=None, **kwargs):
        print("init AbstractGLPlotItem")
        print(f"isinstance(self, GLGraphicsItem): {isinstance(self, GLGraphicsItem)}")
        
        super().__init__(parentItem=parentItem, **kwargs)
        
        
        self._dataset=None
        self._datasetDisplay=None
        self._cacheData = CacheData()
        self.plotdataset_cls=None
        self._viewBox: GL3DViewBox=view_box
        if topology_type=="Mesh":
            self.plotdataset_cls=PlotDatasetMesh
        elif topology_type=="FlatMesh":
            self.plotdataset_cls=PlotDatasetFlatMesh
            
        self.extra_opts={"clipToView" : False,
                         
                         }
        print(f"hasattr(self, ): {hasattr(self, "setProperty")}")
        """
        self.setProperty('xViewRangeWasChanged', False)
        
        self.setProperty('yViewRangeWasChanged', False)
        self.setProperty('zViewRangeWasChanged', False)
        self.setProperty('styleWasChanged', True)  # force initial update
        """
       # self.setParentItem(parentItem)


    def viewTransform(self):
        view = self.getViewBox()
        if view is None:
            return None
        if hasattr(view, 'implements') and view.implements('ViewBox'):
            return self.itemTransform(view.innerSceneItem())[0]
        else:
            return self.sceneTransform()



    def setViewBox(self, vb):
        self._viewBox=vb

    
    #@abc.abstractmethod
    #def _setDataHelper(self, *args, **kwargs):...
    
    #@abc.abstractmethod
    #def setData(self, *args, **kwargs): ...
        
    
    def dataRect(self): 
        return None if self._dataset is None else self._dataset.dataBounds()
    """
    @abc.abstractmethod
    def dataBounds2(self,
                   ax: int,
                   frac=(1,1),
                   orthoRange=None
                   ) -> Tuple[float, float]:...
    """
    def setClipToView(self, state: bool):
        pass
    
    def viewTransformChanged(self):
        self.invalidateBounds()
        #self.prepareGeometryChange()

    def invalidateBounds(self):
        self._cacheData.reset()
        
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):

        # view range has changed; re-plot if needed 
        update_needed = False
        if changed is None or changed[0]: 
            self.setProperty('xViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[1]:
            self.setProperty('yViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True
        if changed is None or changed[2]:
            self.setProperty('zViewRangeWasChanged', True)
            self._datasetDisplay = None
            update_needed = True

        if update_needed:
            self.updateItems(styleUpdate=False)
            
    def updateItems(self, styleUpdate: bool = True):
        dataset = self._getDisplayDataset()
        
        
        if dataset is None:
            raise
            self.hide()
            return
        else:
            self.updateData(**dataset.dataKwargs())
    
    #@classmethod
    #@abstractmethod
    def clipDataFromVRange(cls, dataset, viewrange):
        return dataset
    
    def updateData(self, **kwargs):
        self.invalidateBounds()
        #self.prepareGeometryChange()
        self.informViewBoundsChanged()
      #  self.setData(**kwargs)
        
     #   self.update()
        self.sigPlotChanged.emit(self)
    
    def getData(self):
        dataset = self._getDisplayDataset()
        return (None, None, None) if dataset is None else dataset.data()

        
    def updateData222(self, *args, **kwargs):
        self.invalidateBounds()
        #self.prepareGeometryChange()
        self.informViewBoundsChanged()
        self.update()
        self.sigPlotChanged.emit(self)
    
    def getOriginalDataset(self) -> tuple[None, None] | tuple[np.ndarray, np.ndarray]:
        return self._dataset.data()
    
    def _getDisplayDataset(self):


        if self._dataset is None:
            return None
        # Return cached processed dataset if available and still valid:
        if all((not self._datasetDisplay is None,
                not self.property('xViewRangeWasChanged'),
                not self.property('yViewRangeWasChanged'),
                not self.property('zViewRangeWasChanged')
                )):
            return self._datasetDisplay

        allFinite = self._dataset.allFinite
        data = self._dataset.data()
        
        view = self.getViewBox()
        
        if view is None:
            view_range = None
        else:
            view_range = view.viewRange()  # this is always up-to-date
        if view_range is None:
            view_range = self.viewRect()
 
        if view is None or view.autoRangeEnabled()[0]:
            pass  # no ViewBox to clip to, or view will autoscale to data range.
                
        if not all(allFinite):
            data = self._dataset.data()
        if not view_range is None and all(allFinite):
            data = self.clipDataFromVRange(data, view_range)        

        self._datasetDisplay = self.plotdataset_cls(data, allFinite)
        return self._datasetDisplay
    
    
class CustomGLSurfacePlotItem(AbstractGLPlotItem, opengl.GLSurfacePlotItem):
    """
    **Bases:** :class:`GLMeshItem <pyqtgraph.opengl.GLMeshItem>`
    
    Displays a surface plot on a regular x,y grid
    """
    def __init__(self, x=None, y=None, z=None, colors=None, parentItem=None, **kwds):
        print("CustomGLSurfacePlotItem")
        super().__init__(parentItem=parentItem, x=x, y=y, z=z, colors=colors, **kwds)
        
        
def filter_data_mask22222(x, y, z, limits):
    x_min, x_max = limits[0]
    y_min, y_max = limits[1]
    z_min, z_max = limits[2]
    
    # Create masks for x and y
    x_mask = (x >= x_min) & (x <= x_max)
    y_mask = (y >= y_min) & (y <= y_max)
    
    # Apply masks to x and y
    x_filtered = x[x_mask]
    y_filtered = y[y_mask]
    
    # Apply masks to z (rows correspond to x, columns to y)
    z_filtered = z[x_mask, :][:, y_mask]
    z_filtered = np.clip(z_filtered, z_min, z_max)
    
    return x_filtered, y_filtered, z_filtered

def dataBoundsVectorized(self,
                            frac=(1., 1. ,1),
                            orthoRange=(None, None, None)
                            ) -> Tuple[float, float, float]:

    if not self.visible() or self._dataset is None:
        return [[None, None], [None, None], [None, None]]
    else:
        data = self._dataset.data()
        mask = np.ones(len(data), dtype=bool)

        cache_bounds = {}

        mask_container={}

        for ax in range(2):
            if self._cacheData.cacheFlag(ax) is None and all((self._cacheData.frac(ax) == frac[ax],
                                                                self._cacheData.orthoRange(ax) == orthoRange[ax])):
                cache_bounds[ax] = self._cacheData.bounds(ax)
                mask_container[ax] = self._cacheData.mask(ax)

            if not orthoRange[ax] is None:
                mask = (data[:, ax] >= orthoRange[ax][0]) & (data[:, ax] <= orthoRange[ax][1])
                self._cacheData.setMask(mask, ax)
                mask_container[ax] = mask
            else:
                self._cacheData.setMask(None, ax)
        
        x_filtered = data[0][mask_container[0]] if 0 in mask_container else data[0]
        y_filtered = data[1][mask_container[1]] if 1 in mask_container else data[1]
    
        if self._cacheData.cacheFlag(2) is None and all((self._cacheData.frac(2) == frac[2],
                                                            self._cacheData.orthoRange(2) == orthoRange[2]
                                                            )):
            cache_bounds[2] = self._cacheData.bounds(2)
            
            
        z_filtered = data[2]
        if 0 in mask_container:
            z_filtered = z_filtered[mask_container[0], :]
        
        if 1 in mask_container:
            z_filtered = z_filtered[:, mask_container[1]]

        if not orthoRange[2] is None:
            z_filtered = np.clip(z_filtered, orthoRange[2][0], orthoRange[2][1])

        
        limits = [[bound.min(), bound.max()] if i not in cache_bounds else cache_bounds[i] for i, bound in enumerate((x_filtered, y_filtered, z_filtered))]
        self._cacheData.setBounds(limits)
        return self._cacheData.bounds()

    
def _setDataHelper(self, **kwds):
    colors = kwds.get("colors", None)
    x, y, z = kwds.get("x", None), kwds.get("y", None), kwds.get("z", None)
    if all([ax is None for ax in (x, y, z)]):
        if colors is None:
            kwds["colors"]=colors
            return kwds
    else:
        pos = [ax if not ax is None else getattr(self, f"_{ax_str}") for ax_str, ax in zip("xyz", (x, y, z))]
        self._dataset = self.plotdataset_cls(tuple(pos))
        xyz = self.getData()
        self._datasetDisplay = None

        self._datasetDisplay = None
        kwds.update({ax_str : ax for ax_str, ax in zip("xyz", xyz) if not ax is None})
    if not colors is None:
        kwds["colors"]=colors
    return kwds


def clipDataFromVRange222(cls, data, view_range):
    clipped_dataset=[]
    new_display_z=None
    x, y, z = data
    for ax, arr in enumerate((x,y,z)):
        if ax < 2:
            mask_ax = ([ax] >= view_range[ax][0]) & (arr <= view_range[ax][1])
            clipped_dataset.append(arr[mask_ax])
            if ax == 0:
                new_display_z = arr[mask_ax,:]
            else:
                new_display_z = arr[:,mask_ax]
        else:
            new_display_z = np.clip(new_display_z, view_range[ax][0], view_range[ax][1])
            clipped_dataset.append(new_display_z)

    return tuple(clipped_dataset)


@classmethod
def clipDataFromVRange(cls, data, view_range):
    x_min, x_max = view_range[0]
    y_min, y_max = view_range[1]
    z_min, z_max = view_range[2]
    
    x, y, z = data

    x_mask = (x >= x_min) & (x <= x_max)
    y_mask = (y >= y_min) & (y <= y_max)
    

    x_filtered = x[x_mask]
    y_filtered = y[y_mask]
            
    z_filtered = z[x_mask, :][:, y_mask]
    z_filtered = np.clip(z_filtered, z_min, z_max)

    return x_filtered, y_filtered, z_filtered


def paint(self):        
    
    # WARNING: can return (None, None, None) if self._getDisplayDataset() returns None
    x, y, z = self.getData()

    
    if not x is None and not np.array_equal(self._x, x):

        if self._x is None or len(x) != len(self._x):
            self._vertexes = None
        self._x = x
    
    if not y is None and not np.array_equal(self._y, y):
        if self._y is None or len(y) != len(self._y):
            self._vertexes = None
        self._y = y


    if not z is None and not np.array_equal(self._z, z):

        if not self._x is None and z.shape[0] != len(self._x):
            raise Exception('Z values must have shape (len(x), len(y))')
        if not self._y is None and z.shape[1] != len(self._y):
            raise Exception('Z values must have shape (len(x), len(y))')
        self._z = z
        if not self._vertexes is None and self._z.shape != self._vertexes.shape[:2]:
            self._vertexes = None
    
    
    if self._z is None:
        return
    
    updateMesh = False
    newVertexes = False
    
    ## Generate vertex and face array
    if self._vertexes is None:
        newVertexes = True
        self._vertexes = np.empty((self._z.shape[0], self._z.shape[1], 3), dtype=np.float32)
        self.generateFaces()
        self._meshdata.setFaces(self._faces)
        updateMesh = True
    
    if newVertexes or (not x is None and not np.array_equal(self._x, x)):
        if x is None:
            if self._x is None:
                x = np.arange(self._z.shape[0])
            else:
                x = self._x
        self._vertexes[:, :, 0] = x.reshape(len(x), 1)
        updateMesh = True
    
    if newVertexes or (not y is None and not np.array_equal(self._y, y)):
        if y is None:
            if self._y is None:
                y = np.arange(self._z.shape[1])
            else:
                y = self._y
        self._vertexes[:, :, 1] = y.reshape(1, len(y))
        updateMesh = True
    
    if newVertexes or (not z is None and not np.array_equal(self._z, z)):
        self._vertexes[...,2] = self._z
        updateMesh = True
            
    self.blockUpdates(True)
    if updateMesh:
        self._meshdata.setVertexes(self._vertexes.reshape(self._vertexes.shape[0]*self._vertexes.shape[1], 3))
        self.meshDataChanged()
    self.blockUpdates(False)
    super().paint()



@classmethod
def insert_axis_and_interpolate(cls, arr, axis_values, data, axis=0):
    """Insert missing axis_values into arr, interpolate in data accordingly."""
    arr = np.asarray(arr)
    data = np.asarray(data)
    sort_idx = np.argsort(arr)
    arr = arr[sort_idx]
    data = np.take(data, sort_idx, axis=axis)

    for val in axis_values:
        if val < arr[0] or val > arr[-1] or np.any(np.isclose(arr, val)):
            continue
        idx = np.searchsorted(arr, val)
        a0, a1 = arr[idx - 1], arr[idx]
        t = (val - a0) / (a1 - a0)

        if axis == 0:
            d0, d1 = data[idx - 1], data[idx]
            d_new = (1 - t) * d0 + t * d1
        else:
            d0, d1 = data[:, idx - 1], data[:, idx]
            d_new = (1 - t) * d0 + t * d1

        arr = np.insert(arr, idx, val)
        data = np.insert(data, idx, d_new[np.newaxis, :] if axis == 0 else d_new[:, np.newaxis], axis=axis)

    return arr, data       


def setData(self, x=None, y=None, z=None, colors=None):
    print("setData")
    if not x is None:
        if self._x is None or len(x) != len(self._x):
            self._vertexes = None
        self._x = x
    
    if not y is None:
        if self._y is None or len(y) != len(self._y):
            self._vertexes = None
        self._y = y
    
    if not z is None:
        if not self._x is None and z.shape[0] != len(self._x):
            raise Exception('Z values must have shape (len(x), len(y))')
        if not self._y is None and z.shape[1] != len(self._y):
            raise Exception('Z values must have shape (len(x), len(y))')
        self._z = z
        if not self._vertexes is None and self._z.shape != self._vertexes.shape[:2]:
            self._vertexes = None
    
    if colors is not None:
        self._colors = colors
        self._meshdata.setVertexColors(colors)
    
    if self._z is None:
        return
    
    updateMesh = False
    newVertexes = False
    
    ## Generate vertex and face array
    if self._vertexes is None:
        newVertexes = True
        self._vertexes = np.empty((self._z.shape[0], self._z.shape[1], 3), dtype=np.float32)
        self.generateFaces()
        self._meshdata.setFaces(self._faces)
        updateMesh = True
    
    if newVertexes or x is not None:
        if x is None:
            if self._x is None:
                x = np.arange(self._z.shape[0])
            else:
                x = self._x
        self._vertexes[:, :, 0] = x.reshape(len(x), 1)
        updateMesh = True
    
    if newVertexes or y is not None:
        if y is None:
            if self._y is None:
                y = np.arange(self._z.shape[1])
            else:
                y = self._y
        self._vertexes[:, :, 1] = y.reshape(1, len(y))
        updateMesh = True
    
    if newVertexes or z is not None:
        self._vertexes[...,2] = self._z
        updateMesh = True
    
    if all((not self._x is None, not self._y is None, not self._z is None)):
        pos = [getattr(self, f"_{ax_str}") for ax_str, ax in zip("xyz", (x, y, z))]
        self._dataset = self.plotdataset_cls(tuple(pos))
    self._datasetDisplay=None
    self.blockUpdates(True)
    self.updateItems(styleUpdate=self.property('styleWasChanged'))
    self.blockUpdates(False)

    if updateMesh:
        self._meshdata.setVertexes(self._vertexes.reshape(self._vertexes.shape[0]*self._vertexes.shape[1], 3))
        self.meshDataChanged()
        
    
    #self.updateItems(styleUpdate=self.property('styleWasChanged'))
    #self.blockUpdates(False)
    #self.informViewBoundsChanged()

    #   self.sigPlotChanged.emit(self)
    
    
def generateFaces(self):
    cols = self._z.shape[1]-1
    rows = self._z.shape[0]-1
    faces = np.empty((cols*rows*2, 3), dtype=np.uint32)
    rowtemplate1 = np.arange(cols).reshape(cols, 1) + np.array([[0, 1, cols+1]])
    rowtemplate2 = np.arange(cols).reshape(cols, 1) + np.array([[cols+1, 1, cols+2]])
    for row in range(rows):
        start = row * cols * 2 
        faces[start:start+cols] = rowtemplate1 + row * (cols+1)
        faces[start+cols:start+(cols*2)] = rowtemplate2 + row * (cols+1)
    self._faces = faces

def create_pos():
    X = np.arange(-5, 5, 0.5)
    Y = np.arange(-3, 3, 0.5)
    X_mat, Y_mat = np.meshgrid(X, Y, indexing="xy")
    R = np.sqrt(X_mat**2 + Y_mat**2)

    m1 = 1
    
    Z_mat = m1 * np.sin(R) 
    return X_mat, Y_mat, Z_mat, m1

class CustomColorMap(pg.ColorMap):
    def __init__(self, colourmap_style: str):
        pos = np.linspace(0, 1, 500)          
        try:
            colourmap = pg.colormap.get(colourmap_style)
        except:
            print(f"{colourmap_style} is not in pyqtgraph, using default inferno")
            colourmap = pg.colormap.get("inferno")

        colors = colourmap.map(pos, mode='byte')        
        super().__init__(pos=pos, color=colors, mode='byte')

def create_objects(X_mat, Y_mat, Z_mat, m1) -> AbstractGLPlotItem:

    
    X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
    
    pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
    """
    scatter = GLScatterPlotItem.GLScatterPlotItem(pos=pos_scatter,
                                                  glOptions='opaque',
                                                  color=(1, 0, 1, 1)
                                                  )
    """
    scatter=None
    pos_surface = X_mat[0], Y_mat[:,1], Z_mat.T
    colormap = CustomColorMap("inferno")
    
    z_norm = (Z_mat - Z_mat.min()) / (Z_mat.max() - Z_mat.min())
    colors = colormap.map(np.linspace(0, 1, 500), mode='byte')
    # Map normalized Z to RGB colors


    
    surface = CustomGLSurfacePlotItem(x=pos_surface[0],
                                                  y=pos_surface[1],
                                                  z=pos_surface[2],
                                                  glOptions='opaque',
                                                  shader="shaded",
                                                  colors=colors
                                                  )
    
    print(f"surface._x: {surface._x}")
    return scatter, surface




class Window(QtWidgets.QMainWindow):
    def __init__(self, *args):
        super().__init__(*args)
        
        self.gl_view = opengl.GLViewWidget()
        """
        worldRange=[[0.,1.], [0.,1.],[0.,1.]]

        self.viewBox = GLViewBox.GLViewBox(worldRange=worldRange)
        
        self.gl_view = GLPlotWidget.GLPlotWidget(
                                            # worldRange=worldRange,
                                             #worldRange=np.array([[0.,1.]]*3)
                                             viewBox=self.viewBox
                                             )

        """
   #     self.central_layout = QtWidgets.QVBoxLayout(self)
    #    self.central_layout.addWidget(self.gl_view)
        self.setCentralWidget(self.gl_view)
        self.gl_view.opts["distance"]=5
        self.gl_view.update()
        self.setCentralWidget(self.gl_view)
        self.setGeometry(0, 0, 500, 300)
        self.show()
        self.X_mat, self.Y_mat, self.Z_mat, self.m1 = create_pos()
        self.scatter, self.surface = create_objects(self.X_mat, self.Y_mat, self.Z_mat, self.m1)    
        
        self.gl_view.addItem(self.surface)
        
        
def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
   # win.timer.start(750)
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


import sys
from PySide6 import QtWidgets, QtGui, QtCore

windows = []

def open_new_window(title):
    w = QtWidgets.QMainWindow()
    w.setWindowTitle(title)
    w.setCentralWidget(QtWidgets.QLabel(f"This is {title}"))
    w.resize(400, 300)
    w.show()
    windows.append(w)

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        container = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(container)
        btn1 = QtWidgets.QPushButton("Button 1")
        btn2 = QtWidgets.QPushButton("Button 2")
        btn3 = QtWidgets.QPushButton("Button 3")
        layout.addWidget(btn1)
        layout.addWidget(btn2)
        layout.addWidget(btn3)
        btn1.clicked.connect(lambda: open_new_window("Window from Button 1"))
        btn2.clicked.connect(lambda: open_new_window("Window from Button 2"))
        btn3.clicked.connect(lambda: open_new_window("Window from Button 3"))
        self.setCentralWidget(container)

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    app.setFont(QtGui.QFont("Neue Haas Grotesk"))

    mainWin = MainWindow()
    mainWin.show()
    windows.append(mainWin)

    sys.exit(app.exec())



from __future__ import annotations 
from typing import TYPE_CHECKING

from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GLViewBox import GLViewBox
import importlib
from pprint import pprint   
from OpenGL import GL
from OpenGL.GL import shaders
import numpy as np

from pyqtgraph.Qt import QtGui, QT_LIB
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

__all__ = ['GLLinePlotItem']
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from pyqtgraph.Qt import QT_LIB
import importlib
import pyqtgraph as pg
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DViewBox
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DViewBox
from pyqtgraph import opengl
import sys
import cProfile
import sys
import pstats
from pyqtgraph import ButtonItem, icons
QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

from PySide6.QtTest import QTest
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph.opengl import GLLinePlotItem, GLScatterPlotItem
from pyqtgraph import Transform3D
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DLinePlotDataItem


class CustomObject(GLLinePlotItem):
    def __init__(self, new_items=[], **kwargs):
        self.items=[]
        self.recursive=False
        super().__init__(**kwargs)
        for i in new_items:
            self.addItem(i)

        

    def addItem(self, item):
        self.items.append(item) 
     #   item.setParentItem(self)
        
    def paint(self):
      #  if self.pos is None:
          # return
        self.setupGLState()

        lines = []
      #  color = []
        print(len(self.items))
        colors = []
        for line in self.items:
            
            new_pos =  line.mapToParent(line.pos.T).T
            print("\n")
            line.parent
            pprint(line.transform())
            parent = line.parentItem()
            if parent is not None:
                if not parent.transform().isIdentity():
                    new_pos = parent.transform().map(line.pos.T).T
                    #line.applyTransform(parent.transform(), local=True)

            

            lines.append(new_pos)
      #      colors.append(line.color.tolist())
            colors = colors  + line.color.tolist()

        if len(lines) > 0:  
            self.pos = np.ascontiguousarray(np.vstack(lines))
            self.setData(pos=self.pos)
            
        #    self.color = np.ascontiguousarray(colors)

        super().paint()
        
    def setData(self, **kwds):
        """
        Update the data displayed by this item. All arguments are optional; 
        for example it is allowed to update vertex positions while leaving 
        colors unchanged, etc.
        
        ====================  ==================================================
        **Arguments:**
        ------------------------------------------------------------------------
        pos                   (N,3) array of floats specifying point locations.
        color                 (N,4) array of floats (0.0-1.0) or
                              tuple of floats specifying
                              a single color for the entire item.
        width                 float specifying line width
        antialias             enables smooth line drawing
        mode                  'lines': Each pair of vertexes draws a single line
                                       segment.
                              'line_strip': All vertexes are drawn as a
                                            continuous set of line segments.
        ====================  ==================================================
        """
        args = ['pos', 'color', 'width', 'mode', 'antialias']
        for k in kwds.keys():
            if k not in args:
                raise Exception('Invalid keyword argument: %s (allowed arguments are %s)' % (k, str(args)))
        if 'pos' in kwds:
            pos = kwds.pop('pos')
            self.pos = np.ascontiguousarray(pos, dtype=np.float32)
        if 'color' in kwds:
            color = kwds.pop('color')
            if isinstance(color, np.ndarray):
                color = np.ascontiguousarray(color, dtype=np.float32)
            self.color = color
        for k, v in kwds.items():
            setattr(self, k, v)

        if self.mode not in ['line_strip', 'lines']:
            raise ValueError("Unknown line mode '%s'. (must be 'lines' or 'line_strip')" % self.mode)

        self.vbos_uploaded = False

    def upload_vbo(self, vbo, arr):
        if arr is None:
            vbo.destroy()
            return
        if not vbo.isCreated():
            vbo.create()
        vbo.bind()
        vbo.allocate(arr, arr.nbytes)
        vbo.release()

class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
        
        worldRange=[[0., 2.],
                    [0., 2.],
                    [0., 2.]
                    ]

        y_offsets = np.linspace(0, 1, 4)  # 4 horizontal lines

        lines = []
        for y in y_offsets:
            lines.append([0, y, 0])   # start point
            lines.append([0.5, y, 0])   # end point
            

        lines = np.array(lines)  # shape: (8, 3)

        
        
        lines1, lines2 = lines.copy(), lines.copy()
        

        self.gl_widget = opengl.GLViewWidget()
        
        lines1[0, :] = lines1[0, :] 
        
        lines2[:, 0] = lines2[:,0]/5 - 0.5

        N = lines.shape[0]
        self.line1 = GLLinePlotItem(pos = lines1,
                                    color=np.tile([1.0, 0.0, 0.0, 1.0], (N, 1),),
                                    mode="lines"
                                    )

        self.line2 = GLLinePlotItem(pos = lines2,
                                    color=np.tile([1., 0.0, 1., 1.0], (N, 1),),
                                    mode="lines"
                                    )
        
        self.group = CustomObject(mode="lines", )

        
        self.group.addItem(self.line1)
        self.group.addItem(self.line2)
        
        
        self.line2.setParentItem(self.line1)
        
        
        self.gl_widget.addItem(self.group)
        
        
        
       

        pos = np.array([0, 0, 0])
        colors=["white", "yellow", "cyan"]
        for k in range(3):
            
            pos = np.zeros(3)
            pos[k] = 1 

            
            for i in range(21):
                if i % 2 == 0:
                    gl_item = opengl.GLTextItem(pos=pos * i / 10, text=f"{i/10}", color=colors[k])
                    self.gl_widget.addItem(gl_item), #ignoreBounds=True)
                pos = np.zeros(3)
                pos[k] = 1 
                
        self.setCentralWidget(self.gl_widget)
        self.showMaximized()
        
        
        
        #self.line1.translate(0,dy=1, dz=0, local=False)
        self.line1.update()
        
        self.update()
        
    def update_timer(self):
        self.line1.translate(0,dy=0.5, dz=0, local=False)
        self.update()

        

def main():
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    timer = QtCore.QTimer()
    timer.timeout.connect(win.update_timer)
    timer.start(300)


    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from ...data_classes.classes import AbstractDataClass, VolatilityData, VolVector, Surface, Points, Slice
    from instruments.utils import InstrumentManager
    from instruments.instruments import Spot, Option, Future

from PySide6 import QtWidgets, QtCore
from pyvol_terminal.settings import utils as settings_utils
from datetime import datetime
import numpy as np
from PySide6.QtCore import Qt
import QuantLib as ql
from pyvol_terminal.data_classes import builders as builders_data_classes
from ...quantities import engines
import math
from ..abstract_classes import ABCMainViewQTableWidget


class MainView(ABCMainViewQTableWidget):
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 data_container=None,
                 tick_engine_manager=None,
                 **kwargs):
        super().__init__()
        self.instrument_manager=instrument_manager
        self.options_container=instrument_manager.options_instrument_container
        self.data_container=data_container
        self.column_items=[]
        if tick_engine_manager is None:
            self.tick_engine_manager = engines.TickEngineManager("Strike",
                                                                 "Expiry",
                                                                 "Implied Volatility"
                                                                 )

        else:
            self.tick_engine_manager=tick_engine_manager
        
        self.row_labels=["Style", "Call/Put", "Direction", "Strike", "Expiry", "Model", "Price"]
        self.data_container = self._initDataClassContainer(**kwargs)

        self.spot_object: Spot = list(self.instrument_manager.spot_instrument_container.objects.values())[0]
        
        self.initSettingsTable()
        self._initConstantQLObject()
        
        self.check_until_ready()
        
    def check_until_ready(self):
        xyz = self.data_container["mid"].vol_vector.data()
        if not math.isnan(self.spot_object.mid) and np.isnan(xyz[2]).sum()==0:
            print("Condition met, initializing...")
            print(xyz[2])
            print(np.isnan(xyz[2]).sum()    )
            self.initQLObjects()
        else:
            QtCore.QTimer.singleShot(500, self.check_until_ready)
        
    def _initDataClassContainer(self,       
                                vol_vect_container: VolVector|None=None,
                                default_nTicks: Tuple[int, int]=(30,30),
                                **kwargs
                                ) -> Dict[str, VolatilityData]:
        if vol_vect_container is None:
            return {}
        else:
            return builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container, default_nTicks)

    def initSettingsTable(self):
        self.table = QtWidgets.QTableWidget()
        self.table.setRowCount(6)
        self.table.setColumnCount(3)
        self.table.verticalHeader().setVisible(False)
        self.table.horizontalHeader().setVisible(False)
        
        for idx, row_label in enumerate(self.row_labels):
            self.table.setItem(idx, 0, QtWidgets.QTableWidgetItem(row_label))
    
    def initModelOptionsTable(self):
        self.table = QtWidgets.QTableWidget()
        self.table.setRowCount(6)
        self.table.setColumnCount(3)
        
        
    def _initConstantQLObject(self):
        expiration_dt = [datetime.fromtimestamp(ts) for ts in self.data_container["mid"].surface.data()[1]]
        self.expiration_dates_ql = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt]
        self.calculation_date = ql.Date(datetime.now().day, datetime.now().month, datetime.now().year)
        ql.Settings.instance().evaluationDate = self.calculation_date

    def initQLObjects(self):
        v0, kappa, theta, rho, sigma = 0.01, 0.2, 0.02, -0.75, 0.5
        dfs=[1.]
        futures_exp_dates = [self.calculation_date]
        
        spot = self.spot_object.mid

        for futures_object in self.instrument_manager.futures_instrument_container.objects.values():
            exp_ts = futures_object.expiry
            exp_dt = datetime.fromtimestamp(exp_ts)
            exp_ql = ql.Date(exp_dt.day, exp_dt.month, exp_dt.year)
            dfs.append(futures_object.mid / spot)
            futures_exp_dates.append(exp_ql)
        
        data_container = self.data_container["mid"]
        xyz = np.column_stack(data_container.vol_vector.data())



        xyz = MainView.getRectStrikesExpiries(xyz)

        strikes = np.unique(xyz[:, 0])
        expiries = np.unique(xyz[:, 1])
        vol_grid = xyz[:, 2].reshape(len(strikes), len(expiries))
        print(vol_grid)
        expiration_dt = [datetime.fromtimestamp(ts) for ts in expiries]
        self.expiration_dates_ql = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt]
        

        
        implied_vols = ql.Matrix(vol_grid.shape[0], vol_grid.shape[1])
        
        
        for i, _ in enumerate(strikes):
            for j, _ in enumerate(expiries):
                implied_vols[i][j] = vol_grid[i][j]

        
        calendar = ql.UnitedStates(ql.UnitedStates.NYSE)
        day_count = ql.Actual365Fixed()
        black_var_surface = ql.BlackVarianceSurface(self.calculation_date,
                                                    calendar,
                                                    self.expiration_dates_ql,
                                                    strikes.tolist(),
                                                    implied_vols,
                                                    day_count
                                                    )
                
        
        strikes_grid = np.arange(strikes[0], strikes[-1], 10)
        expiry = 0.6
        implied_vols = [black_var_surface.blackVol(expiry, s)
                        for s in strikes_grid] # can interpolate here

        print(implied_vols) 
        
        _ = dfs.pop(1)
        _ = futures_exp_dates.pop(1)
        
        
        self.calendar = ql.NullCalendar()
        curve = ql.DiscountCurve(futures_exp_dates, dfs, ql.Actual360())
        self.interest_ts = ql.YieldTermStructureHandle(curve)
        self.dividend_ts = ql.YieldTermStructureHandle(ql.FlatForward(self.calculation_date, ql.QuoteHandle(ql.SimpleQuote(0.0)), ql.Actual365Fixed())),  # No dividend yield
        
        
        
        process = ql.HestonProcess(self.interest_ts, self.dividend_ts,
                                ql.QuoteHandle(ql.SimpleQuote(spot)),
                                v0, kappa, theta, sigma, rho)

        self.model = ql.HestonModel(process)
        self.engine = ql.AnalyticHestonEngine(self.model)
    
    def calibrate(self):
        heston_helpers = []
        for j, (date, ts) in enumerate(zip(self.expiration_dates_ql, self.data_container.domain["mid"].y_vect)):
            for i, s in enumerate(self.data_container.domain["mid"].x_vect):
                t = (date - self.calculation_date)
                p = ql.Period(t, ql.Days)
                sigma = self.data_container.raw
                helper = ql.HestonModelHelper(p, self.calendar, self.spot_object.mid, s,
                                              ql.QuoteHandle(ql.SimpleQuote(sigma)),
                                              self.interest_ts,
                                              self.dividend_ts)
                helper.setPricingEngine(self.engine)
                heston_helpers.append(helper)
        lm = ql.LevenbergMarquardt(1e-8, 1e-8, 1e-8)
        self.model.calibrate(heston_helpers, lm, ql.EndCriteria(500, 50, 1.0e-8, 1.0e-8, 1.0e-8))
        theta, kappa, sigma, rho, v0 = self.model.params()
        print((theta, kappa, sigma, rho, v0))        

    def _internal_update_view(self):
        pass



    @classmethod
    def getRectStrikesExpiries(cls, xyz) -> Tuple[np.ndarray, ...]:
        print(len(np.unique(xyz[:, 0])))
        strikes, expiries, vol = xyz[:,0], xyz[:,1], xyz[:,2]
        exp = np.unique(expiries)
        unique, counts = np.unique(strikes, return_counts=True)
        k = unique[counts >= exp.size]
        
        xyz = xyz[np.isin(xyz[:, 0], k)]
        xyz = xyz[np.isin(xyz[:, 1], exp)]
        
        return xyz
    
    @classmethod
    def drop_duplicate_strike_exp(cls, arr: np.ndarray) -> np.ndarray:
        col01 = arr[:, :2]
        _, inv, counts = np.unique(col01, axis=0, return_inverse=True, return_counts=True)
        mask = counts[inv] == 1
        return arr[mask]


import sys
import numpy as np
from PySide6 import QtWidgets, QtGui, QtCore
import pyqtgraph.opengl as gl
from OpenGL.GL import *


class GroupedGLTextItem(gl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, parent=None):
        super().__init__()
        self._positions = np.empty((0, 3))
        self._texts = []
        self._color = QtGui.QColor(255, 255, 255)
        self._font = QtGui.QFont("Helvetica", 12)

    def setData(self, positions, texts, color=QtGui.QColor(255, 255, 255)):
        self._positions = np.array(positions)
        self._texts = texts
        self._color = color
        self.update()

    def paint(self):
        view = self.view()
        if view is None:
            return

        self.setupGLState()
        cam = view.cameraParams()

        modelview = glGetDoublev(GL_MODELVIEW_MATRIX)
        projection = glGetDoublev(GL_PROJECTION_MATRIX)
        width, height = view.width(), view.height()
        viewport = (0, 0, width, height)

        painter = QtGui.QPainter(view)
        painter.setPen(self._color)
        painter.setFont(self._font)

        for pos, text in zip(self._positions, self._texts):
            screen_pos = self.__project(pos, modelview, projection, viewport)
            if screen_pos is not None:
                painter.drawText(int(screen_pos.x()), int(screen_pos.y()), text)

        painter.end()

    def __project(self, obj_pos, modelview, projection, viewport):
        obj_vec = np.append(np.array(obj_pos), [1.0])

        view_vec = np.matmul(modelview.T, obj_vec)
        proj_vec = np.matmul(projection.T, view_vec)

        if proj_vec[3] == 0.0:
            return QtCore.QPointF(0, 0)

        proj_vec[0:3] /= proj_vec[3]

        return QtCore.QPointF(
            viewport[0] + (1.0 + proj_vec[0]) * viewport[2] / 2,
            viewport[1] + (1.0 + proj_vec[1]) * viewport[3] / 2
        )


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Optimized 3D Text Rendering")
        self.setGeometry(100, 100, 800, 600)

        central_widget = QtWidgets.QWidget()
        self.setCentralWidget(central_widget)
        layout = QtWidgets.QVBoxLayout(central_widget)

        self.view = gl.GLViewWidget()
        self.view.setCameraPosition(distance=20)
        layout.addWidget(self.view)

        # Add grid for context
        grid = gl.GLGridItem()
        grid.setSize(10, 10)
        grid.setSpacing(1, 1)
        self.view.addItem(grid)

        # Create grouped text item
        self.text_item = GroupedGLTextItem()
        positions = [[0, 0, 0], [2, 2, 0], [-2, 1, 1]]
        texts = ["Origin", "Top Right", "Left Up"]
        self.text_item.setData(positions, texts, color=QtGui.QColor("yellow"))
        self.view.addItem(self.text_item)


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from ...gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DSurfacePlotItem, GL3DScatterPlotItem
    from .pyvol_GL3DGraphicsItems import ABCPyVolPlotItemMixin

    
import numpy as np
from pyqtgraph import opengl 
from PySide6 import QtCore, QtGui
from . import utils as utils_volatility_surface
from ... import utils
from pyvol_terminal import misc_widgets
from ...gl_3D_graphing.graphics_items import GL3DGraphicsItems
from pyqtgraph import functions as fn
import math
import pyqtgraph as pg
from ...gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget
from .pyvol_GL3DGraphicsItems import PyVolGL3DScatterPlotItem, PyVolGL3DSurfacePlotItem, PyVolLineItem
import traceback
import time


class PyVolGL3DViewWidget(GL3DViewWidget):    
    rcMapCoordsSig = QtCore.Signal(float, float, float)
    hoverSurfaceSig = QtCore.Signal(float, float, float)
    viewChangedSig = QtCore.Signal(object)
    intenseInteractionEnded = QtCore.Signal()
    
    def __init__(self,
                 spot_text=None,
                 padding=None,
                 queue_interaction: bool=True,
                 **kwargs
                 ):
        if padding is None:
            padding = [0, 0, 0]
        
        self.spot_objects = kwargs.pop("spot_objects", None)
        parent=kwargs.pop("parent", None),
        print(kwargs.keys())
        super().__init__(padding=padding, **kwargs)
        self._mv_handle_interaction_state=None
        self.spot_text=spot_text
        self._cartesian_view_coords = np.array([0., 0., 0.])
        self.queue_interaction=queue_interaction

            
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setMouseTracking(True)
        self.setAttribute(QtCore.Qt.WA_Hover)

        self.show_spot_text=True
        self.surface_flag=True
        self.scatter_flag=True
        self.surface_visibility={}
        self.spot_text_dict={}
        self.displayed_price_types=[]
        self.surface_plotitems: Dict[str, PyVolGL3DSurfacePlotItem]={}
        self.scatter_plotitems: Dict[str, PyVolGL3DScatterPlotItem]={}
        self._mouseDrag=False
        self._normalising_data={}
        self.plot_items_by_attr = {}
        self.plotted_price_types=[]
        self._displayed_price_type_counter={}
        self.cross_hairs_on=False
        self.text_color = (0, 0, 0, 255)
        self.bg_color = (255, 255, 255, 200)
        self.cross_hairs_enabled=True
        self._mouseZoomFactor=0.999
        
        near_clip, far_clip = self.compute_optimal_clipping()
        
        distance = 4
        azimuth = -90
        elevation = 30
        
        
        self._intense_interacting = False
        
        
        self.opts['near'] = near_clip
        self.opts['far'] = far_clip
        self.opts["distance"] = distance
        self.opts['azimuth'] = azimuth
        self._azimuth_offset = 90
        self._elevation_offset = 90
        
        self.opts["elevation"] = elevation
        self.opts["center"] = QtGui.QVector3D(0, 1, 0)

        self._spherical_bounds = {"azimuth" : [-90, 0],
                                  "elevation" : [0, 90]}
        self.prev_euler_coords = {"distance" : distance,
                                  "azimuth" : azimuth,
                                  "elevation" : elevation}

        self._bounds_cartesian = [[0, 1], [0, 1], [0, 1]]

        
        self.first_plot=True
        self.plot_interaction_buffer=[] 
        self.any_valid_surface=False
        self.price_updated_callbacks=[]
        self._plot_items: List[PyVolLineItem]=[]

        self.prev_x, self.prev_y, self.prev_z = None, None, None
        self.addPlots_callbacks=[]
        self.RemovePlots_callbacks=[]
        
        self._last_update_time = 0
        self._update_interval = 50  # ms between updates
        self._pending_update = False

        self.hoverSurfaceSig.connect(self._update_crosshairs)
        self._cartesian_view_coords = self.cameraPositionVector()
        
        self.scroll_delay = None
        if self.queue_interaction:
            self.scroll_delay = 200
            self._mouseScroll=False
            self.scroll_timer = QtCore.QTimer()
            self.scroll_timer.setSingleShot(True)
            self.scroll_timer.timeout.connect(self._processMouseZoom)
            
        self.setup_mouse_interaction()
        self._init_crosshairs()
        
        self.top_surface: PyVolGL3DSurfacePlotItem=None
        self.add_legend()    

    def wheelEvent(self, ev):
        if self.queue_interaction:
            self._mouseScroll=True
            self._internalCheckIntenseInteraction()
            self.scroll_timer.start(self.scroll_delay)
        return super().wheelEvent(ev)
    
    def _processMouseZoom(self):
        self._mouseScroll=False
        self._internalCheckIntenseInteraction()
  
    def plotItem(self,
                 internal_id
                 ) -> ABCPyVolPlotItemMixin:
        return self.id_map.get(internal_id, None)
    
    def append_addPlots_callbacks(self, callback):
        self.addPlots_callbacks.append(callback)

    def append_removePlots_callbacks(self, callback):
        self.RemovePlots_callbacks.append(callback)

    def repaint_spot(self):
        painter = QtGui.QPainter(self)
        painter.setRenderHints(QtGui.QPainter.Antialiasing | QtGui.QPainter.TextAntialiasing)
        
        margin = 10
        spot_text=str(self.spot_text)
        text_rect = painter.boundingRect(self.rect(), 
                                         QtCore.Qt.AlignTop | QtCore.Qt.AlignRight | QtCore.Qt.TextDontClip, 
                                         spot_text
                                        )
        text_rect.moveTopRight(self.rect().topRight() - QtCore.QPoint(margin, -margin))
        
        painter.setBrush(QtGui.QColor(*self.bg_color))
        painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 200), 1))
        painter.drawRect(text_rect.adjusted(-5, -2, 5, 2))
        painter.setPen(QtGui.QColor(*self.text_color))
        painter.drawText(text_rect, QtCore.Qt.AlignCenter, spot_text)
        painter.end()
        
    @property
    def top_surface(self) -> PyVolGL3DSurfacePlotItem:
        return self._top_surface
    
    @top_surface.setter
    def top_surface(self, surfaceItem: PyVolGL3DSurfacePlotItem):
        self.line_xz.setParentItem(surfaceItem)
        self.line_yz.setParentItem(surfaceItem)
        self._top_surface = surfaceItem

    def get_top_price_type(self):
        if "ask" in self.surface_plotitems:
            self.top_surface = self.surface_plotitems["ask"]
        elif "mid" in self.surface_plotitems:
            self.top_surface = self.surface_plotitems["mid"]
        elif "bid" in self.surface_plotitems:
            self.top_surface = self.surface_plotitems["bid"]
        
    def update_surface_state(self, gl_plot_item: PyVolGL3DSurfacePlotItem):
        self.surface_visibility[gl_plot_item.id()] = gl_plot_item.valid_values
        self.any_valid_surface = any(self.surface_visibility.values())
                    
    def plotitem_from_attr(self, px_type, plot_type):
        return self.plot_items_by_attr[px_type][plot_type]
    
    def _add_to_plot_data_item_dicts(self, item: PyVolLineItem):
        self._normalising_data[item.id()]=False
        if item.px_type in self._displayed_price_type_counter:
            self._displayed_price_type_counter[item.px_type]+=1
        else:
            self._displayed_price_type_counter[item.px_type]=1

    def addItem(self, item, ignoreBounds=False, **kwargs):
        if isinstance(item, PyVolLineItem):
            self._plot_items.append(item)
        
        if isinstance(item, (PyVolGL3DSurfacePlotItem, PyVolGL3DScatterPlotItem)):
            container = self.surface_plotitems if isinstance(item, PyVolGL3DSurfacePlotItem) else self.scatter_plotitems
            container[item.px_type] = item
            self._add_to_plot_data_item_dicts(item)
            self.legend.add_legend_item(item.px_type, item.color)
            self.id_map[item.id()] = item
            
            if isinstance(item, PyVolGL3DSurfacePlotItem):
                self.get_top_price_type()
                item.sigPlotChanged.connect(self.update_surface_state)

        super().addItem(item, ignoreBounds=ignoreBounds, **kwargs)
            
    def removeItem(self, item: ABCPyVolPlotItemMixin):
        if isinstance(item, PyVolLineItem):
            self._plot_items.remove(item)

        if isinstance(item, PyVolGL3DSurfacePlotItem):
            del self.surface_plotitems[item.px_type]
            self.get_top_price_type()
        if isinstance(item, PyVolGL3DScatterPlotItem):
            del self.scatter_plotitems[item.px_type]
            
        if isinstance(item, GL3DGraphicsItems.GL3DScatterPlotItem) or isinstance(item, GL3DGraphicsItems.GL3DScatterPlotItem):
            self.legend.remove_legend_item(item.px_type)
           # del self.plot_items_by_attr[item.px_type][item.item_type]
            self._displayed_price_type_counter[item.px_type]-=1
            if self._displayed_price_type_counter[item.px_type]==0:
                del self._displayed_price_type_counter[item.px_type]
        return super().removeItem(item)
    
    def add_legend(self, legend=None):
        if not legend is None:
            self.legend=legend
        else:
            self.legend=misc_widgets.Legend()
        
    def remove_legend(self):
        if not self.legend is None:
            self.removeItem(self.legend)
            self.legend=None

    def remove_spot_text(self):
        self.show_spot_text = False
        self.spot_text = ""
        self.update()
        
    def toggle_crosshairs(self, check, on_off):
        print(check)
        if on_off == "On":
            enable=True
        else:
            enable=False
        self.cross_hairs_enabled=enable
        if not enable:
            self.line_xz.hide()
            self.line_yz.hide()
            
    def restore_spot_text(self, text=""):
        self.show_spot_text = True
        self.spot_text = text
        self.update()

    def paintGL3(self, *args, **kwargs):
        super().paintGL(*args, **kwargs)
        if self.show_spot_text and self.spot_text:
            painter = QtGui.QPainter(self)
            self.repaint_spot()
            painter.end()

    def compute_optimal_clipping(self):
        largest_point_distance = np.sqrt(2)
        near_clip = max(0.0001, self.opts['distance'] - 1.01 * largest_point_distance) 
        far_clip = self.opts['distance'] 
        return near_clip, far_clip

    def _init_crosshairs(self):
        self.line_yz = None
        self.line_xz = None
        self.cross_hairs_on = False
        line_xz, line_yz = [np.column_stack([np.linspace(0,1,60)]*3)]*2 , np.column_stack([np.linspace(0,1,60)]*3)
        
        self.line_xz = PyVolLineItem(pos=line_xz,
                                     color=(1, 1, 1, 1),
                                     width=2,
                                     mode='line_strip',
                                     antialias=True
                                     )
        self.line_yz = PyVolLineItem(pos=line_yz,
                                    color=(1, 1, 1, 1),
                                    width=2,
                                    mode='line_strip',
                                    antialias=True
                                    )
        self.line_yz.setGLOptions('translucent')
        self.line_xz.setGLOptions('translucent')
        self.line_yz.hide()
        self.line_xz.hide()

    def setup_mouse_interaction(self):
        self.setMouseTracking(True)
        
        self.mouse_move_timer = QtCore.QTimer(self)
        self.mouse_move_timer.setInterval(50)
        self.mouse_move_timer.timeout.connect(self.process_mouse_move)
        self.mouse_move_timer.start()
        
        self.lastMouseWorldCoords=None
        
    def mousePressEvent(self, event):
        if not self.vb.mouse_state.keyInteracting():
            match event.buttons().value:    
                case 2:
                    self.process_right_click(event.pos().x(), event.pos().y())
        super().mousePressEvent(event)

    def process_right_click(self, screen_x, screen_y):
        self.makeCurrent() 
        world_x, world_y, world_z = utils_volatility_surface.map_2D_coords_to_3D(self, screen_x, screen_y)
        if 0 <= world_x <= 1 and 0 <= world_y <= 1 and 0 <= world_z <= 1:        
            local_coords = self.vb.childGroup.mapFromView(pg.Vector(world_x, world_y, world_z))

            self.prev_x = local_coords.x()
            self.prev_y = local_coords.y()
            self.prev_z = local_coords.z()
            self.rcMapCoordsSig.emit(self.prev_x,
                                     self.prev_y,
                                     self.prev_z)
    
    def hide_crosshairs(self):
        self.cross_hairs_on=False
        self.line_yz.hide()
        self.line_xz.hide()

    def leaveEvent(self, event):
        self.lastMouseWorldCoords=None
        self.hide_crosshairs()
        super().leaveEvent(event)

    def mouseMoveEvent(self, ev):
        self.lastMouseWorldCoords = ev.x(), ev.y()
        if self.queue_interaction and ev.buttons() & QtCore.Qt.LeftButton:
            self._mouseDrag=True
            self._internalCheckIntenseInteraction()
        return super().mouseMoveEvent(ev)
        
    def _delayed_view_update(self):
        if self._pending_update:
            self._emit_view_changed()
            
    def _emit_view_changed(self):
        self._last_update_time = QtCore.QTime.currentTime().msecsSinceStartOfDay()
        self._pending_update = False
        self.viewChangedSig.emit(self)
        
    def _internalCheckIntenseInteraction(self):
        state = self._mouseDrag or self._mouseScroll
        if state != self._intense_interacting:
            self._intense_interacting=state
            self.processIntenseInteraction("view", state)
        
    def mouseReleaseEvent(self, event):
        if self.queue_interaction and self._mouseDrag:
            self._mouseDrag=False
            self._internalCheckIntenseInteraction()
        self._mouseDrag=False
        return super().mouseReleaseEvent(event)
    
    def _update_crosshairs(self, x, y, z):
        vect = pg.Vector(x, y, z)
        vect = self.top_surface.mapFromView(vect)
        
        x, y = vect.x(), vect.y()
        
        if not self.cross_hairs_on:
            self.cross_hairs_on = True
            self.line_yz.show()
            self.line_xz.show()

        line_xz = utils.calculate_xy_lines(*self.top_surface.getValues(),
                                            x_fixed=x,
                                            y_fixed=y,
                                            const_axis=True
                                            )
  
        
        line_yz = utils.calculate_xy_lines(*self.top_surface.getValues(),
                                            x_fixed=x,
                                            y_fixed=y,
                                            const_axis=False
                                            )
        
        line_xz[:,2]*=1.05
        line_yz[:,2]*=1.05
        
        self.line_xz.setData(pos=line_xz)
        self.line_yz.setData(pos=line_yz)
    
    def process_mouse_move(self):
        if (self.cross_hairs_enabled
            and self.any_valid_surface
            and not self.lastMouseWorldCoords is None
            ):
            self.makeCurrent() 
            world_x, world_y, world_z = utils_volatility_surface.map_2D_coords_to_3D(self, *self.lastMouseWorldCoords)
            if 0 <= world_x <= 1 and 0 <= world_y <= 1 and 0 <= world_z <= 1:
                self.hoverSurfaceSig.emit(world_x, world_y, world_z)
            else:
                self.cross_hairs_on=False
                self.line_xz.hide()
                self.line_yz.hide()
            self.lastMouseWorldCoords = None

    def get_legend(self):
        return self.legend

    def get_displayed_price_types(self):
        return self._displayed_price_type_counter


    def zoom(self, rMultiplier):
        distance = self.opts['distance'] * rMultiplier
        opts = {"distance" : distance,
                "elevation" : self.opts["elevation"],
                "azimuth" : self.opts["azimuth"]
                }
        if self._outsideBox(opts):
            self.opts["distance"] = distance
            self.update()
            return True
        else:
            return False

    def cameraPositionVector(self, opts=None) -> np.ndarray[float]:
        if opts is None:
            opts = self.opts

        azimuth = opts["azimuth"] + self._azimuth_offset
        elevation = self._elevation_offset - opts["elevation"]
        self._cartesian_view_coords[0] = opts["distance"] * math.sin(math.radians(elevation)) * math.cos(math.radians(azimuth))
        self._cartesian_view_coords[1] = opts["distance"] * math.sin(math.radians(elevation)) * math.sin(math.radians(azimuth))
        self._cartesian_view_coords[2] = opts["distance"] * math.cos(math.radians(elevation))
        return self._cartesian_view_coords

    def _checkCartesianBounds(self, coords):
        return [self._bounds_cartesian[i][0] <= coords[i] < self._bounds_cartesian[i][1] for i in range(3)]
    
    def _outsideBox(self, opts):
        if any([self._bounds_cartesian[idx][1] <= pos for idx, pos in enumerate(self.cameraPositionVector(opts))]):
            return True
        else:
            return False
        
    def _validAngle(self, azimuth, elevation):
        if (self._spherical_bounds["azimuth"][0] <= azimuth <= self._spherical_bounds["azimuth"][1]
            and self._spherical_bounds["elevation"][0] <= elevation <= self._spherical_bounds["elevation"][1]):
            return True
        else:
            return False
    
    def _validView(self, opts):
        cartesian = self.cameraPositionVector(opts)
        
        if any([pos < self._bounds_cartesian[idx][0] or pos > self._bounds_cartesian[idx][1] for idx, pos in enumerate(cartesian)]):
            return False
        else:
            return True
        
    def cameraPositionNumpy(self):
        p = self.cameraPosition()
        return np.array((p.x(), p.y(), p.z()))
    
    def processIntenseInteraction(self, *args, **kwargs): ...


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from axis.axis_utils import TickEngineManager, TickEngine
    
import pyqtgraph as pg
from PySide6 import QtWidgets
from typing import List
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
from ... import misc_widgets
from .. import subplots
from pyvol_terminal.axis import widgets as axis_widgets
from typing import Dict
from pyvol_terminal.data_classes import builders as builders_data_classes
from pyvol_terminal.axis import axis_utils as utils_axis


class CustomPlotitem(subplots.SubPlot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plot_items={}
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.getViewBox().disableAutoRange(0)
        self.getViewBox().disableAutoRange(1)
        
        self._plot_items: List[pg.PlotDataItem] = []
        self.x_buffer = 0.1
        self.y_buffer = 0.1

    def add_legend(self, legend):
        self.legend = legend
    
    def addPlots(self, px_type, value, line, scatter):
        self.plot_items[px_type]["line"].update({value : line})
        self.plot_items[px_type]["scatter"].update({value : scatter})
        super().addItem(line)
        super().addItem(scatter)
        line.sigPlotChanged.emit(self.calculate_limits)
            
    def removeItem(self, item):
        if isinstance(item, pg.PlotDataItem) and item in self._plot_items:
            try:
                item.sigPlotChanged.disconnect(self.calculate_limits)
            except (RuntimeError, TypeError):
                pass 
            self._plot_items.remove(item)
        result = super().removeItem(item)
        self.calculate_limits()
    
    def _set_limits(self, x_min, x_max, y_min, y_max):
        if x_min == x_max:
            x_offset = 0.5 
        else:
            x_offset = (x_max - x_min) * self.x_buffer        
        if y_min == y_max:
            y_offset = 0.5
        else:
            y_offset = (y_max - y_min) * self.y_buffer
        self.getViewBox().setXRange(x_min - x_offset, x_max + x_offset)
        self.getViewBox().setYRange(y_min - y_offset, y_max + y_offset)    
        self.update()
        
    def calculate_limits(self):
        x_data = []
        y_data = []

        for plot_item in self._plot_items:
            x_min, x_max, y_min, y_max = plot_item.get_limits()
            x_data.append(x_min)
            x_data.append(x_max)
            y_data.append(y_min)
            y_data.append(y_max)
        if any(~np.isnan(x_data)):
            self.x_min, self.x_max = np.nanmin(x_data), np.nanmax(x_data)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y_data)):
            self.y_min, self.y_max = np.nanmin(y_data), np.nanmax(y_data)
        else:
            self.y_min, self.y_max = np.nan, np.nan
        
        if (not np.isnan(self.x_min)
            and not np.isnan(self.x_max)
            and not np.isnan(self.y_min)
            and not np.isnan(self.y_max)):
            if len(self._plot_items) == 1:
                self._set_limits(x_min, x_max, y_min, y_max)
            else:
                x_range, y_range = self.getPlotItem().getViewBox().viewRange()
                if (x_min < x_range[0] or x_max > x_range[1] or y_min < y_range[0] or y_max > y_range[1]):
                    self._set_limits(x_min, x_max, y_min, y_max)
        

class MainView(QtWidgets.QSplitter):
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 interpolation_config,
                 tick_engine_manager: TickEngine,
                 slice_direction: str,
                 update_timer=0,
                 **kwargs 
                ):
        super().__init__(QtCore.Qt.Horizontal)
        self.displayed_domains=[]
        self.container_scatter=[]
        self.container_line=[]
        self.data_container=[]
        base_domain, domain = builders_data_classes.create_domains(instrument_manager.options_instrument_container.original_data)
    
        self.metric_axis_engine=utils_axis.MetricAxisEngine(base_domain.strike,
                                                            base_domain.expiry,
                                                            "Strike",
                                                            "Expiry",
                                                            "Implied Volatility")
        self.data_container = builders_data_classes.create_surface_dataclasses(instrument_manager.config["options"]["price_types"], 
                                                                               instrument_manager.options_instrument_container.original_data,
                                                                               domain,
                                                                               self.metric_axis_engine,
                                                                               instrument_manager.options_instrument_container.get_objects(),
                                                                               interpolation_config)
        tick_engine = getattr(tick_engine_manager, f"{slice_direction}_engine")
        self.x_axis_item = axis_widgets.CustomAxisItem(axis_direction="x",
                                                       tick_engine=tick_engine,
                                                       orientation="bottom") 
        self.y_axis_item = axis_widgets.CustomAxisItem(axis_direction="x",
                                                       tick_engine=tick_engine_manager.z_engine,
                                                       orientation="right") 
        
        axisItems = {"right" : self.y_axis_item, "bottom" : self.x_axis_item}
        self.plot.setAxisItems(axisItems)
        self.plot = CustomPlotitem(axisItems=axisItems)
        self.plot.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Expanding)
                
        self.legend = misc_widgets.Legend()
        self.plot.addLegend(offset=(0, 0))   
        self.init_layout(tick_engine_manager.x_engine)


    def init_layout(self, tick_engine):
        self.setLayout(QtWidgets.QGridLayout(self))
        
        self.layout().addWidget(self.plot, 0, 0)
        self.layout().addWidget(self.legend, 0, 0, alignment=QtCore.Qt.AlignTop | QtCore.Qt.AlignRight)
        
    
    def addPlots(self, px_type):
        for value in self.displayed_domains:
            self.create_plot_objects(px_type, value)
    
    def addPlot(self, px_type, value):
        self.create_plot_objects(px_type, value)
    
    def create_plot_objects(self, px_type, value):
        self.data_container.add_slice(px_type, value)
        slice = self.data_container.slice_container[px_type][value]
        scatter = CustomPlotDataItem(x=slice.x, y=slice.y)
        line = CustomPlotDataItem(x=slice.xi, y=slice.yi)
        self.container_scatter[px_type][value] = scatter
        self.container_line[px_type][value] = line
        self.addPlots(line)
        self.addPlots(scatter)
        line.show()



class CustomPlotWidget(pg.PlotWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.getPlotItem().getViewBox().disableAutoRange(0)
        self.getPlotItem().getViewBox().disableAutoRange(1)
        self._plot_items: List[pg.PlotDataItem] = []
        self.x_buffer = 0.1
        self.y_buffer = 0.1

    def add_curve(self, item):
        self.getPlotItem().addItem(item)
        self._plot_items.append(item)
        self.calculate_limits()  
        item.sigPlotChanged.emit(self.calculate_limits)
            
    def removeItem(self, item):
        if isinstance(item, pg.PlotDataItem) and item in self._plot_items:
            try:
                item.sigPlotChanged.disconnect(self.calculate_limits)
            except (RuntimeError, TypeError):
                pass 
            self._plot_items.remove(item)
        result = super().removeItem(item)
        self.calculate_limits()
    
    def _set_limits(self, x_min, x_max, y_min, y_max):
        if x_min == x_max:
            x_offset = 0.5 
        else:
            x_offset = (x_max - x_min) * self.x_buffer        
        if y_min == y_max:
            y_offset = 0.5
        else:
            y_offset = (y_max - y_min) * self.y_buffer
        self.getPlotItem().getViewBox().setXRange(x_min - x_offset, x_max + x_offset)
        self.getPlotItem().getViewBox().setYRange(y_min - y_offset, y_max + y_offset)    
        self.update()
        
    def calculate_limits(self):
        x_data = []
        y_data = []

        for plot_item in self._plot_items:
            x_min, x_max, y_min, y_max = plot_item.get_limits()
            x_data.append(x_min)
            x_data.append(x_max)
            y_data.append(y_min)
            y_data.append(y_max)
        if any(~np.isnan(x_data)):
            self.x_min, self.x_max = np.nanmin(x_data), np.nanmax(x_data)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y_data)):
            self.y_min, self.y_max = np.nanmin(y_data), np.nanmax(y_data)
        else:
            self.y_min, self.y_max = np.nan, np.nan
        
        if (not np.isnan(self.x_min)
            and not np.isnan(self.x_max)
            and not np.isnan(self.y_min)
            and not np.isnan(self.y_max)):
            if len(self._plot_items) == 1:
                self._set_limits(x_min, x_max, y_min, y_max)
            else:
                x_range, y_range = self.getPlotItem().getViewBox().viewRange()
                if (x_min < x_range[0] or x_max > x_range[1] or y_min < y_range[0] or y_max > y_range[1]):
                    self._set_limits(x_min, x_max, y_min, y_max)
        
class CustomScatterPlotItem(pg.ScatterPlotItem):
    pen_map: Dict[str, pg.mkPen]
    def __init__(self, px_type, *args, **kwargs):
        kwargs["symbol"] = "x"
        kwargs["size"]=10
        kwargs["pxMode"]=True
        self.px_type = px_type
        super().__init__(*args, **kwargs)



class CustomPlotDataItem(pg.PlotDataItem):
    pen_map: Dict[str, pg.mkPen]
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._calculate_limits(*self.getData())
    
    def _calculate_limits(self, x, y):
        if any(~np.isnan(x)):
            self.x_min, self.x_max = np.min(x), np.max(x)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y)):
            self.y_min, self.y_max = np.min(y), np.max(y)
        else:
            self.y_min, self.y_max = np.nan, np.nan
    
    def get_limits(self):
        return self.x_min, self.x_max, self.y_min, self.y_max
    
    def get_xlim(self):
        return self.x_min, self.x_max
    
    def get_ylim(self):
        return self.y_min, self.y_max

    def setData(self, **kwargs):
        self._calculate_limits(kwargs["x"], kwargs["y"])
        super().setData(**kwargs)
        



from PySide6 import QtGui, QtWidgets, QtCore
from . import stylesheets


class OptionMetricCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() | QtCore.Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~QtCore.Qt.ItemIsEditable)
        self.setBackground(QtGui.QBrush("black"))
        self.setForeground(QtGui.QBrush("#fb8b1e"))
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)


class Option(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(self.flags() | QtCore.Qt.ItemIsSelectable)
        self.setFlags(self.flags() & ~QtCore.Qt.ItemIsEditable)
        self.setBackground(QtGui.QBrush("black"))
        self.setForeground(QtGui.QBrush("#fb8b1e"))
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)


class OptionExpiryCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        self.setBackground(QtGui.QBrush("black"))
        self.setForeground(QtGui.QBrush("#fb8b1e"))
        self.setTextAlignment(QtCore.Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

class TableColumnItem(QtWidgets.QTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        self.setBackground(QtGui.QBrush("#232323"))
        self.setForeground(QtGui.QBrush("white"))
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)



class StrikeOptionsComboBox(QtWidgets.QComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.view().setAutoScroll(False)
        self.setStyleSheet(stylesheets.get_settings_stylesheets("StrikeOptionsComboBox"))        
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

    def wheelEvent(self, event):
        if self.view().isVisible():
            super().wheelEvent(event)
            




from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from . import tables, table_builder, extra_widgets, settings
import time
from . import utils as utils_omon
from .tables import OptionTableWidget, StrikeTableColumn
from ..abstract_classes import ABCMainViewQTableWidget

class MainView(ABCMainViewQTableWidget):
    def __init__(self,
                 instrument_manager: InstrumentManager=None,
                 expiry_underlying_map=None,
                 tick_label_engine=None,
                 update_timer=0,
                 queue_interaction=None,
                 metricLabels: List[str]=[],
                 parent=None,
                 **kwargs
                 ):
        super().__init__(parent=parent)
        self.calls_table: OptionTableWidget=None
        self.puts_table: OptionTableWidget=None
        self.options_container=instrument_manager.options_instrument_container
        self.update_timer=update_timer
        self.last_update=time.time()
        self.strike_table_column: StrikeTableColumn=None
        self._subtable_update_callbacks: List[Callable]=[]
        expiry_attributes, total_rows = self.init_expiry_attr()
        
        self.scroll_delay=None
        self.queue_interaction=queue_interaction
        
        if self.queue_interaction:
            self.scroll_delay = 300
            self.scroll_timer = QtCore.QTimer()
            self.scroll_timer.setSingleShot(True)
            self.scroll_timer.timeout.connect(self._scrollFinished)

        self.init_option_tables(instrument_manager, total_rows, expiry_attributes, metricLabels, expiry_underlying_map, tick_label_engine)
        self.init_layout()

    def _internal_update_view(self, *args):
        self.calls_table.update_table()
        self.puts_table.update_table()  
        for callback in self._subtable_update_callbacks:
            callback()
                
    def add_subtable_update_callback(self, callback):
        self._subtable_update_callbacks.append(callback)
        
    def init_expiry_attr(self):
        all_strikes_per_expiry = {expiry : strike_arr for expiry, strike_arr in self.options_container.maps.expiry_strike_map.items()}        
        
        expiry_strike_idx_map = {}
        expiry_idx_map = {}
    
        total_rows=0
        for expiry, strike_arr in all_strikes_per_expiry.items():
            temp_dict = {}
            expiry_idx_map[expiry] = total_rows
            total_rows+=1
            for strike in strike_arr:
                temp_dict[strike] = total_rows
                total_rows+=1
            expiry_strike_idx_map[expiry] = temp_dict
        
        expiry_attributes={}

        for expiry in self.options_container.maps.expiry_strike_map:
            expiry_attribute = utils_omon.ExpiryAttributes(expiry,
                                                           np.array(all_strikes_per_expiry[expiry]),
                                                           np.array(all_strikes_per_expiry[expiry]).size,
                                                           expiry_idx_map[expiry],
                                                           expiry_strike_idx_map[expiry]
                                                           )
            expiry_attributes[expiry] = expiry_attribute
        return expiry_attributes, total_rows

    def init_option_tables(self, instrument_manager, total_rows, expiry_attributes, metric_labels, expiry_underlying_map, tick_label_engine):
        self.strike_table_column = tables.StrikeTableColumn(total_rows, expiry_attributes)   # A QTableWidget object

        table_args = [instrument_manager,
                      metric_labels,
                      expiry_attributes,
                      total_rows,
                      ]

        table_kwargs = {"expiry_underlying_map" : expiry_underlying_map,
                        "tick_label_engine" : tick_label_engine,
                        "parent" : self
                        }
        self.calls_table = tables.OptionTableWidget(*table_args,
                                                    "c",
                                                    **table_kwargs
                                                    )
        self.puts_table = tables.OptionTableWidget(*table_args,
                                                   "p",
                                                   **table_kwargs
                                                   )
        self.strike_table_column.sigRowsUpdated.connect(self._sync_strike_to_options_heights)
        self.strike_table_column.sigRowsUpdated.emit(self.strike_table_column)
        for expiry_attribute in expiry_attributes.values():
            for idx in expiry_attribute.strike_idx_map.values():
                self.strike_table_column.add_row_hidden_callback(self.calls_table.setRowHidden, idx)
                self.strike_table_column.add_row_hidden_callback(self.puts_table.setRowHidden, idx)
        
        for child_table in [self.calls_table, self.puts_table, self.strike_table_column]:
            child_table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
            child_table.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        
        self.link_scrollbars(*[self.calls_table, self.puts_table, self.strike_table_column])
                
        self._link_row_heights(self.strike_table_column,
                               self.calls_table,
                               self.puts_table)
        
        self.calls_table.right_click_menu.setParent(self.parent())
        self.puts_table.right_click_menu.setParent(self.parent())
    
    def match_initial_row_heights(self):
        for i in range(1, self.strike_table_column.rowCount()):
            height = self.strike_table_column.rowHeight(i)
            self.calls_table.setRowHeight(i-1, height)
            self.puts_table.setRowHeight(i-1, height)
    
    def _sync_strike_to_options_heights(self, strike_table):
        for strike_row in range(1, strike_table.rowCount()):
            height = strike_table.rowHeight(strike_row)
            target_row = strike_row - 1  # Adjust for blank row
            self.calls_table.setRowHeight(target_row, height)
            self.puts_table.setRowHeight(target_row, height)    

    def _link_row_heights(self, master, *slaves: Tuple[tables.OptionTableWidget,...]):
        vh = master.verticalHeader()                  
        def _copy_height(logical_row, _old, new):
            for table in slaves:
                table.setRowHeight(logical_row, new)
                
            self.strike_table_column.setRowHeight(logical_row + 1, new)
        vh.sectionResized.connect(_copy_height)
        
        for row in range(master.rowCount()):
            h = master.rowHeight(row)
            for table in slaves:
                table.setRowHeight(row, h)
            self.strike_table_column.setRowHeight(row + 1, h)
                        
    def link_scrollbars(self, *tables):
        for table in tables:
            table.verticalScrollBar().valueChanged.connect(lambda value, current=table: self.sync_scroll(value, current, tables))

    def sync_scroll(self, value, source, tables):
        for table in tables:
            if not table is source:
                table.verticalScrollBar().setValue(value)
        
    def update_spot_text(self):
        pass
        
    def init_layout(self):
        self.setRowCount(1)
        self.setColumnCount(3)
        self.verticalHeader().setVisible(False)
        self.setHorizontalHeaderItem(0, extra_widgets.TableColumnItem("Calls"))
        self.setHorizontalHeaderItem(1, extra_widgets.TableColumnItem("Strike"))
        self.setHorizontalHeaderItem(2, extra_widgets.TableColumnItem("Puts"))
        
        self.setCellWidget(0, 0, self.calls_table)
        self.setCellWidget(0, 1, self.strike_table_column)
        self.setCellWidget(0, 2, self.puts_table)
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)  
        self.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.Fixed)
        self.setColumnWidth(1, self.strike_table_column.columnWidth(0))

        self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.horizontalHeader().setSectionResizeMode(2, QtWidgets.QHeaderView.Stretch)        
        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        
        
    def wheelEvent(self, ev):
        if self.queue_interaction:
            self._mouseScroll=True
            self._internalCheckIntenseInteraction()
            self.scroll_timer.start(self.scroll_delay)
        return super().wheelEvent(ev)
    
    def _scrollFinished(self):
        self._mouseScroll=False
        self._internalCheckIntenseInteraction()

    def _internalCheckIntenseInteraction(self):
        self.processIntenseInteraction("table", self._mouseScroll)
        
    def setColumns(self, columns):
        pass
    
    def currrentColumns(self):
        return self.calls_table.currrentColumns()



import numpy as np
from PySide6 import QtWidgets, QtCore, QtGui
import pyqtgraph as pg
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DGraphicsItems import GL3DSurfacePlotItem, GL3DScatterPlotItem
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox
from pyvol_terminal.gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DViewBox

import sys

def create_pos():
    X = np.arange(-5, 5, 0.5)
    Y = np.arange(-3, 3, 0.5)
    X_mat, Y_mat = np.meshgrid(X, Y, indexing="xy")
    R = np.sqrt(X_mat**2 + Y_mat**2)

    m1 = 2
    
    Z_mat = m1 * np.sin(R) 
    return X_mat, Y_mat, Z_mat, m1

class CustomColorMap(pg.ColorMap):
    def __init__(self, colourmap_style: str):
        pos = np.linspace(0, 1, 500)          
        try:
            colourmap = pg.colormap.get(colourmap_style)
        except:
            print(f"{colourmap_style} is not in pyqtgraph, using default inferno")
            colourmap = pg.colormap.get("inferno")

        colors = colourmap.map(pos, mode='byte')        
        super().__init__(pos=pos, color=colors, mode='byte')

def create_objects(X_mat, Y_mat, Z_mat, m1):

    
    X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
    
    pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
    """
    scatter = GLScatterPlotItem.GLScatterPlotItem(pos=pos_scatter,
                                                  glOptions='opaque',
                                                  color=(1, 0, 1, 1)
                                                  )
    """
    scatter=None
    pos_surface = X_mat[0], Y_mat[:,1], Z_mat.T
    colormap = CustomColorMap("inferno")
    
    z_norm = (Z_mat - Z_mat.min()) / (Z_mat.max() - Z_mat.min())
    colors = colormap.map(np.linspace(0, 1, 500), mode='byte')
    # Map normalized Z to RGB colors
  #  colors = viridis_like.map(z_norm.flatten(), mode='byte')[:, :3]  # shape: (N, 3)\

    #colors = colors.reshape(Z_mat.shape + (3,))  # shape: (rows, cols, 3)

    
    surface = GL3DSurfacePlotItem(x=pos_surface[0],
                                    y=pos_surface[1],
                                    z=pos_surface[2],
                                    glOptions='opaque',
                                    shader="shaded",
                                    colors=colors
                                    )
        
    return scatter, surface




class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)      
        self.X_mat, self.Y_mat, self.Z_mat, self.m1 = create_pos()
        self.scatter, self.surface = create_objects(self.X_mat, self.Y_mat, self.Z_mat, self.m1)    
        worldRange=[[0.,1.], [0.,1.],[0.,1.]]
        
        self.viewBox = GL3DViewBox(worldRange=worldRange)
        ax_orth = worldRange[0][1], worldRange[2][0]
        axisItem2 = GL3DViewBox(1,
                               ax_orth,
                               0,
                               )
        axis_items = [axisItem2]

        self.gl_view = GL3DViewWidget(defaultWorldRange=worldRange,
                                    viewBox=self.viewBox,
                                    axisItems=axis_items,
                                    )
        self.gl_view.opts["distance"] = 10

        self.X_mat, self.Y_mat, self.Z_mat, self.m1 = create_pos()
        self.scatter, self.surface = create_objects(self.X_mat, self.Y_mat, self.Z_mat, self.m1)    
       # import sys
      #  sys.exit()
        
        self.scatter=None
       # self.gl_view.addItem(self.scatter)
        self.gl_view.addItem(self.surface)


        self.setCentralWidget(self.gl_view)
        self.showMaximized()
        


        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_plots)

    def update_plots(self):
        self.m1 += 1
        xmin, xmax = self.X_mat.min() - 1, self.X_mat.max() + 1
        ymin, ymax = self.Y_mat.min() - 1, self.Y_mat.max() + 1
        X = np.arange(xmin, xmax, 0.5)
        Y = np.arange(ymin, ymax, 0.5)
        self.X_mat, self.Y_mat = np.meshgrid(X, Y)
        R = np.sqrt(self.X_mat**2 + self.Y_mat**2)

        
        
        self.Z_mat = self.m1 * np.sin(R) 
        
        X_vec, Y_vec, Z_vec = self.X_mat.flatten(), self.Y_mat.flatten(), self.Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vec))
        pos_surface = self.X_mat[0], self.Y_mat[:,1], self.Z_mat.T
        
        
        
        #AbstractGLPlotItem.AbstractGLPlotItem.setData(self.scatter, pos=new_pos)
        
     #   self.scatter.setData(pos=pos_scatter)
        
        self.surface.setData(x=pos_surface[0],
                             y=pos_surface[1],
                             z=pos_surface[2]
                             )
        """
        self.surface.setData(x=self.surface._x,
                             y=self.surface._y,
                             z=self.surface._z,
                             )
        """

def main():
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    win.timer.start(1000)
    sys.exit(app.exec())
if __name__ == "__main__":
    main()


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.terminal.application import Application
    from pyvol_terminal.interfaces.abstract_classes import ABCInterface
    from .workers import ABCCalibrationWorker
    

from PySide6 import QtCore, QtWidgets
import time
from pyvol_terminal.settings import widgets as widgets_settings
import time
from pyqtgraph import opengl

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)


class InterfaceWindow1(QtWidgets.QMainWindow):
    aboutToCloseSig = QtCore.Signal(QtWidgets.QMainWindow)
    
    def __init__(self,
                 primary_instrument,
                 calibration_worker: ABCCalibrationWorker,
                 **interface_configs
                 ):
        super().__init__()
        self.primary_instrument=primary_instrument

        widget_class = interface_configs["widget_class"]
        
        self.interface: ABCInterface = widget_class(parent=self, **interface_configs)

        self.calibration_worker=calibration_worker
        self.calibration_worker.calibratedSignal.connect(self.interface.main_view.update_view)

        if not self.calibration_worker.isRunning():
            self.calibration_worker.start_worker()
        
        self._initLayout()
        self.showMaximized()
        
    def _initLayout(self):     
        self.stacked_interfaces = QtWidgets.QStackedLayout()
        self.stacked_interfaces.addWidget(self.interface)

        self.setCentralWidget(QtWidgets.QWidget())
        self.central_layout = QtWidgets.QVBoxLayout()
        self.central_layout.addLayout(self.stacked_interfaces)
        self.centralWidget().setLayout(self.central_layout)
    
    def closeEvent(self, event):
        self.calibration_worker.calibratedSignal.disconnect(self.interface.main_view.update_view)
        self.aboutToCloseSig.emit(self)
        super().closeEvent(event)
        
        
class SubWindow(QtWidgets.QMainWindow):
    def __init__(self, widget, geometry: QtCore.QRect=None, adjustToContents: bool=False, flags=None, **kwargs):
        super().__init__(**kwargs)
        self.setCentralWidget(widget)
        if not geometry is None:
            self.setGeometry(geometry)
        if flags:
            self.setWindowFlags(flags)
        self._adjustToContents = adjustToContents
        if adjustToContents:
            QtCore.QTimer.singleShot(0, self.adjustToContents)
        self.show()
    
    def adjustToContents(self):
        self.adjustSize()
        self.setFixedSize(self.size())

class SubWindowTable(SubWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.adjustWindowToTable()

    def adjustWindowToTable(self):
        width = self.centralWidget().verticalHeader().width() + \
                self.centralWidget().horizontalHeader().length() + \
                self.centralWidget().frameWidth() * 2
        
        height = self.centralWidget().horizontalHeader().height() + \
                 self.centralWidget().verticalHeader().length() + \
                 self.centralWidget().frameWidth() * 2
        self.adjustSize()




import sys
from PySide6 import QtCore, QtWidgets, QtGui
import pyqtgraph as pg
from pyqtgraph import opengl


class CustomTransform(QtGui.QTransform):
    
    def __init__(self, parent=None):
        super().__init__()
        self._vertical_shear = 0.0 
        self._horizontal_shear = 0.0
        self.updateTransform()

    def getVerticalShear(self):
        return self._vertical_shear
    
    def setVerticalShear(self, shear):
        if self._vertical_shear != shear:
            self._vertical_shear = shear
            self.updateTransform()

    def getHorizontalShear(self):
        return self._horizontal_shear

    def setHorizontalShear(self, shear):
        if self._horizontal_shear != shear:
            self._horizontal_shear = shear
            self.updateTransform()

    def updateTransform(self):
        self.setMatrix(1.0, self._horizontal_shear, 0.0,
                      self._vertical_shear, 1.0, 0.0,
                      0.0, 0.0, 1.0)

    verticalShear = QtCore.Property(float, getVerticalShear, setVerticalShear)
    horizontalShear = QtCore.Property(float, getHorizontalShear, setHorizontalShear)

class GraphicsObject(QtWidgets.QGraphicsObject):
    def __init__(self, parent=None):
        super().__init__(parent)


class ItemGroup(GraphicsObject):

    def __init__(self, parent=None):
        super().__init__(parent)

    def addItem(self, item):
        if item and item.parentItem() != self:
            item.setParentItem(self)
        self.normaliseToparent()


    def paint(self, painter, option, widget=None):
        pass

    def boundingRect(self):
        return self.childrenBoundingRect()
    
    def normaliseToparent(self):
        rect = self.boundingRect()
        
        tr = self.transform()
        tr.scale(1/rect.width(), 1/rect.height())
        tr.translate(-rect.x(), -rect.y())
        self.update()



class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("ItemGroup with QOpenGLWidget Example")

        self.scene = QtWidgets.QGraphicsScene()
        
        self.view = QtWidgets.QGraphicsView(self.scene)
        self.view.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        
        self.openGLWidget = pg.Qt.QtWidgets.QOpenGLWidget()
        self.view.setViewport(self.openGLWidget)
        self.group = ItemGroup()
        
        tr = CustomTransform()
        self.group.setTransform(tr)
        
        self.scene.addItem(self.group)
        
        self.rect1 = QtWidgets.QGraphicsRectItem(0, 0, 50, 50)
        self.rect1.setBrush(QtGui.QColor("crimson"))
        self.group.addItem(self.rect1)

        self.rect2 = QtWidgets.QGraphicsRectItem(1200, 1200, 50, 50)
        self.rect2.setBrush(QtGui.QColor("steelblue"))
        self.group.addItem(self.rect2)
        
        self.group.setFlag(QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.group.setFlag(QtWidgets.QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        

        self.setCentralWidget(self.view)
        self.resize(400, 300)
        self.showMaximized()
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        rect = self.scene.itemsBoundingRect()
        print(rect)
        self.update()
        rect = self.scene.itemsBoundingRect()
        
        print(rect)

        
       # self.timer.start(1500)
        
    def update(self):
        self.rect1.setRect(2 * 1200, 2*1200, 50, 50)
        

        
        

def main():
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    sys.exit(app.exec())

if __name__ == '__main__':
    main()



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.utils import InstrumentManager
    from pandas import DataFrame
    from engines.surface_engines import AbstractSurfaceEngine
    from data_classes.classes import BaseDomain, Domain, OptionChain, VolatilityData, VolVector
    
from PySide6 import QtCore
import asyncio
import queue
import time
import numpy as np

class WebsocketWorker(QtCore.QThread):
    update_signal = QtCore.Signal(list, bool)  

    def __init__(self,parallel_type=None, ws_transport_method=None, websocket=None,
                 start_ws_func_name="", q=None, multiple_levels=False,
                 qt_timer_seconds=2, bulk_response=False):
        super().__init__()
        self._is_running = True  
        self.stop_flag=False
        self.parallel_type=parallel_type
        self._should_stop = False
        self.price_generator = websocket
        self.start_ws_func_name=start_ws_func_name
        self.all_response = []
        self.bulk_response = bulk_response
        self.q=q
        self.loop=None
        self.task=None
        self.qt_timer_seconds=qt_timer_seconds
        self.multiple_levels=multiple_levels
        self.ws_transport_method=ws_transport_method
        self.generator_call = getattr(self.price_generator, start_ws_func_name)            
        self._should_stop = False
        
        if self.ws_transport_method == "queue":
            self.queue_timer = QtCore.QTimer()
            self.queue_timer.timeout.connect(self.get_queue)
            self.queue_timer.start(self.qt_timer_seconds * 1000)
    
    def get_queue(self):
        all_responses = []
        while True:
            try:
                response = self.q.get_nowait()
                all_responses.append(response)
            except queue.Empty:
                break 
        if len(all_responses) > 0:
            self.update_signal.emit(all_responses, True)
            
    def run_threading(self,):
        self.generator_call()

    async def run_async(self,):
        async for message in self.generator_call():
            if self._should_stop:  
                break          
            self.update_signal.emit([message], False)

    def run(self):
        match self.parallel_type:
            case "async":
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)
                self.loop.run_until_complete(self.run_async())
                self.loop.close()
            case "threading":
                self.run_threading()

                                
    def stop(self):
        self._should_stop = True
        self._is_running = False
        """
        if self.loop:
            self.loop.close()
        """
        self.quit()
        self.wait()

        


def get_updater_dict():
    return {""}


class PriceProcessor(QtCore.QObject):
    processedSig = QtCore.Signal()
    
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 data_processing_config):
        super().__init__()
        self.instrument_manager=instrument_manager
        self.last_buffer_responses={}
        self.last_process_update=time.time()
        self.data_processing_config=data_processing_config
        websocket_json_format = data_processing_config["websocket_json_format"]
        self.ws_instrument_key = websocket_json_format["instrument_key"]
        self.ws_bid_key = websocket_json_format["bid_key"]
        self.ws_ask_key = websocket_json_format["ask_key"]
        self.ws_timestamp_key = websocket_json_format["timestamp_key"]
        self.last_buffer_update_time=time.time()
        self._last_update_checker_timer=time.time()
        
        if "buffer" in data_processing_config:
            if data_processing_config["buffer"]:
                if "timer_process_data" in data_processing_config:
                    self.timer_process_data=data_processing_config["timer_process_data"]
                else:
                    self.timer_process_data=2
                self._updater_method = self.buffer_updater
            else:
                self.timer_process_data=None
                self._updater_method=self.single_update
        else:
            self.timer_process_data=None
            self._updater_method=self.single_update
    
    def update_interface_signal():
        pass
    
    def _update_displays(self):
        self.update_interface_signal.emit()
    
    def single_update(self, websocket_response):
        self.update_price(websocket_response)
        self.processedSig.emit()
    
    def update_price(self, websocket_response):
        ticker = websocket_response[self.ws_instrument_key]    
        timestamp = websocket_response[self.ws_timestamp_key]
        bid = websocket_response[self.ws_bid_key]
        ask = websocket_response[self.ws_ask_key]
        self.instrument_manager.update_price(ticker, timestamp, bid, ask)  # Ensures sister option is ITM/OTM
        
        if time.time() - self._last_update_checker_timer > 20:
            for option_object in self.instrument_manager.all_instrument_objects.values():
                if time.time() - option_object.last_update_time > 20:
                    option_object.update(ticker, timestamp, bid, ask)

    def check_enough_time(self):
        if time.time() - self.last_process_update > self.timer_process_data:            
            return True
        else:
            return False
    
    def buffer_updater(self, websocket_response):
        self.bulk_response(websocket_response)
        if self.check_enough_time():
            self.update_price_with_buffer()
        self.processedSig.emit()
    
    
    QtCore.Slot()
    def update(self, websocket_response):
        self._updater_method(websocket_response)
    
    def bulk_response(self, websocket_responses):
        for websocket_response in websocket_responses:
            ticker = websocket_response[self.ws_instrument_key]
            self.last_buffer_responses[ticker] = websocket_response
                
    def update_response_buffer(self, websocket_response):
        ticker = websocket_response[self.ws_instrument_key]
        self.last_buffer_responses[ticker] = websocket_response

    def update_price_with_buffer(self):
        for ticker, websocket_response in self.last_buffer_responses.copy().items():
            self.update_price(websocket_response)
            del self.last_buffer_responses[ticker]
        self.last_buffer_responses.clear()
        self.last_process_update=time.time()



class SurfaceEvaluate(QtCore.QThread):
    calibratedSignal = QtCore.Signal()
    
    def __init__(self,
                 raw_options_container: Dict[str, OptionChain],
                 update_timer: float=0.,
                 engine_container: Dict[str, SurfaceEvaluate]={},
                 vol_vect_container: Dict[str, Dict[str, VolVector]]={},
                 **kwargs,
                 ):
        super().__init__()
        self.raw_options_container=raw_options_container
        self.vol_vect_container=vol_vect_container
        self._update_timer=update_timer
        self.surface_engine_container: Dict[str, AbstractSurfaceEngine]=engine_container
        self._last_calibration=time.time()
        self._last_price_process=time.time()
        self._timer = None
    
    @QtCore.Slot()
    def price_process_response(self):
        self._last_price_process=time.time()
    
    def calibrate_surfaces(self):
        if self._last_price_process > self._last_calibration:
            for px_type, surface_engine in self.surface_engine_container.items():
                for vol_vect_container in self.vol_vect_container.values():
                    raw_options = self.raw_options_container[px_type]
                    vol_vector = vol_vect_container[px_type]
                    vol_vector.update(**raw_options.getDataKwargs())
                
                    surface_engine.calibrate(*vol_vector.data())

            self._last_calibration=time.time()
            self.calibratedSignal.emit()
    
    def add_vol_vect_container(self, name, vol_vect_container):
        self.vol_vect_container[name]=vol_vect_container
        
    def remove_vol_vect_container(self, name):
        del self.vol_vect_container[name]

    def run(self):
        self.timer = QtCore.QTimer()
        self.timer.moveToThread(self)
        self.timer.timeout.connect(self.calibrate_surfaces)
        self.timer.start(self._update_timer*1000)
        self.exec()    


    
    def stop(self):
        if self._timer:
            self._timer.stop()
            self._timer.deleteLater()